#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports an extended CMU Seasons dataset to the kapture format.
It creates one kapture dataset for each of the 24 slices, based on the txt files

This is an extended version of the dataset published in
T. Sattler, W. Maddern, C. Toft, A. Torii, L. Hammarstrand, E. Stenborg, D. Safari,
M. Okutomi, M. Pollefeys, J. Sivic, F. Kahl, T. Pajdla.
Benchmarking 6DOF Outdoor Visual Localization in Changing Conditions.

Conference on Computer Vision and Pattern Recognition (CVPR) 2018

This dataset is in turn based on the CMU Visual Localization dataset described here:
Hernan Badino, Daniel Huber, and Takeo Kanade.
The CMU Visual Localization Data Set.
http://3dvis.ri.cmu.edu/data-sets/localization, 2011.

The Extended CMU Seasons dataset uses a subset of the images provided in the CMU Visual Localization dataset.
It uses images taken under a single reference condition (sunny + no foliage),
captured at 17 locations (referred to as slices hereafter), to represent the scene.
For this reference condition, the dataset provides a reference 3D model reconstructed using Structure-from-Motion.
The 3D model consequently defines a set of 6DOF reference poses for the database images.
In addition, query images taken under different conditions at the 17 slices are provided.
Reference poses for around 50% of these images are included in this dataset,
in addition to the reference sequence as described above.

https://www.visuallocalization.net/datasets/

Intrinsics

Please refer to the intrinsics.txt file for a description of how the camera intrinsics are specified.
Note that non-linear distortion are present in the images.
The intrinsics.txt file contains information about this distortion.

Database Lists

For each slice, we provide a text file with the list of database images for that slice.
The text file stores a line per database image. Here is an example from slice 2:
img_00122_c0_1303398475046031us.jpg 0.636779 0.569692 0.379586 -0.354792 -85.392177 55.210379 -2.980700
Here, img_00122 indicates that this image is the 122 image in the dataset.
The infix _c1_ indicates that camera 1 was used to capture the image.
1283347879534213us is the timestamp of the capture time.
The seven numbers after the image name is the camera pose:
the first four are the components of a rotation quaternion, corresponding to a rotation R,
and the last three are the camera center C.
The rotation R corresponds to the first 3x3 subblock of the corresponding camera matrix,
and the camera center C is related to the fourth column t of the camera matrix according to C = -R^T * t,
where R^T denotes the transpose of R.

Query images

A list of all query images for each slice is provided in the test-images-sliceX.txt files.
"""

import logging
import os
import os.path as path
import quaternion
import argparse
import re
import numpy as np
from typing import List, Optional, Tuple

# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir, float_array_or_none
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.algo.merge_keep_ids import merge_keep_ids

logger = logging.getLogger('Extended_CMU_Seasons')
ECMU_IMAGE_PATTERN = r'img_(?P<image_number>\d+)_(?P<camera_id>\w+)_(?P<timestamp>\d+)us\.jpg'


def import_extended_cmu_seasons_intrinsics(intrinsics_file_path: str) -> kapture.Sensors:
    """
    Read and convert intrinsics file
    Format: [Camera ID] [Distortion model] [image width] [image height] [fx] [fy] [cx] [cy] [k1] [k2] [p1] [p2]

    :param intrinsics_file_path: path to the CMU intrinsics file
    :return: kapture cameras
    """
    cameras = kapture.Sensors()
    with open(intrinsics_file_path) as fin:
        table = fin.readlines()
        # remove comment lines
        table = (l1 for l1 in table if not l1.startswith('#'))
        # remove empty lines
        table = (l2 for l2 in table if l2.strip())
        # trim trailing EOL
        table = (l3.rstrip("\n\r") for l3 in table)
        # split space
        table = (re.split(r'\s+', l4) for l4 in table)
        # remove empty split
        table = ([s for s in l5 if s] for l5 in table)

    for camera_id, distortion_model, *camera_params in table:
        cameras[camera_id] = kapture.Camera(distortion_model, list(camera_params))

    return cameras


def _parse_image_name(image_name: str, compiled_image_pattern) -> Tuple[Optional[int], Optional[str]]:
    matches = compiled_image_pattern.match(image_name)
    if not matches:
        logger.warning(f"Error matching image_name in {image_name}")
        return None, None

    matches = matches.groupdict()

    # image_number = str(matches['image_number'])
    camera_id = str(matches['camera_id'])
    timestamp = int(matches['timestamp'])
    return timestamp, camera_id


def import_extended_cmu_seasons_images(image_list_file_path: str) -> Tuple[kapture.RecordsCamera, kapture.Trajectories]:
    """
    Read image list, name.jpg or name.jpg qw qx qy qz cx cy cz

    :param image_list_file_path: path to the image list file
    :return: kapture images and trajectories
    """

    records_camera = kapture.RecordsCamera()
    trajectories = kapture.Trajectories()

    # name.jpg qw qx qy qz cx cy cz
    # or
    # name.jpg
    with open(image_list_file_path) as fin:
        table = fin.readlines()
        # remove comment lines
        table = (line for line in table if not line.startswith('#'))
        # remove empty lines
        table = (line for line in table if line.strip())
        # trim trailing EOL
        table = (line.rstrip("\n\r") for line in table)
        # split space
        table = (re.split(r'\s+', line) for line in table)
        # remove empty split
        table = ([s for s in line if s] for line in table)

    image_pattern = re.compile(ECMU_IMAGE_PATTERN)
    for line in table:
        image_name = line[0]
        timestamp, camera_id = _parse_image_name(image_name, image_pattern)
        if camera_id is None or timestamp is None:
            continue

        records_camera[(timestamp, camera_id)] = image_name
        if len(line) > 1:  # also contains trajectory
            qw, qx, qy, qz, cx, cy, cz = line[1:]
            quaternion_array = float_array_or_none([qw, qx, qy, qz])
            assert quaternion_array is not None
            center_array = float_array_or_none([cx, cy, cz])
            assert center_array is not None
            rotation = quaternion.from_float_array(quaternion_array)
            # C = -R^T * t -> t = -R * C
            translation = np.matmul(quaternion.as_rotation_matrix(rotation),
                                    -1 * np.array(center_array, dtype=np.float))
            pose = kapture.PoseTransform(r=rotation, t=translation)
            trajectories[(timestamp, camera_id)] = pose

    # if no trajectories were added (query), prefer None
    if not trajectories:
        trajectories = None

    return records_camera, trajectories


def _add_images_from_folder(images_folder_path: str, kapture_data: kapture.Kapture):
    file_list = [os.path.relpath(os.path.join(dirpath, filename), images_folder_path)
                 for dirpath, dirs, filenames in os.walk(images_folder_path)
                 for filename in filenames
                 if filename.endswith('us.jpg')]

    already_added_images = set(image_name for _, _, image_name in kapture.flatten(kapture_data.records_camera))
    image_pattern = re.compile(ECMU_IMAGE_PATTERN)
    for image_name in file_list:
        if image_name not in already_added_images:
            timestamp, camera_id = _parse_image_name(image_name, image_pattern)
            if camera_id is None or timestamp is None:
                continue
            kapture_data.records_camera[(timestamp, camera_id)] = image_name


def import_extended_cmu_seasons(cmu_path: str,
                                top_kaptures_path: str,
                                slice_range: List[int],
                                import_all_files: bool = False,
                                force_overwrite_existing: bool = False,
                                images_import_method: TransferAction = TransferAction.skip) -> None:
    """
    Import extended CMU data to kapture. Will make training and query kaptures for every CMU slice.

    :param cmu_path: path to the top directory of the CMU dataset files
    :param top_kaptures_path: top directory for the kaptures to create
    :param slice_range: range of CMU slices to import
    :param import_all_files: if Tre, will import all files
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    """

    os.makedirs(top_kaptures_path, exist_ok=True)

    cameras = import_extended_cmu_seasons_intrinsics(path.join(cmu_path, 'intrinsics.txt'))

    for slice_n in slice_range:
        # prepare paths
        slice_path = os.path.join(cmu_path, f'slice{slice_n}')
        training_images_path = os.path.join(slice_path, 'database')
        query_images_path = os.path.join(slice_path, 'query')
        gt_trajectories_path = os.path.join(slice_path, f'ground-truth-database-images-slice{slice_n}.txt')
        query_image_list = os.path.join(slice_path, f'test-images-slice{slice_n}.txt')
        query_gt_path = os.path.join(slice_path, 'camera-poses')
        query_gt_list = [os.path.join(query_gt_path, f) for f in os.listdir(query_gt_path)]

        # Import training images
        kapture_training_path = path.join(top_kaptures_path, f'slice{slice_n}', "mapping")
        delete_existing_kapture_files(kapture_training_path, force_overwrite_existing)
        training_records_camera, training_trajectories = import_extended_cmu_seasons_images(gt_trajectories_path)
        training_kapture = kapture.Kapture(sensors=cameras,
                                           records_camera=training_records_camera,
                                           trajectories=training_trajectories)
        if import_all_files:
            _add_images_from_folder(training_images_path, training_kapture)
        kapture_to_dir(kapture_training_path, training_kapture)
        # finally import images
        if images_import_method != TransferAction.skip:
            filename_list = [f for _, _, f in kapture.flatten(training_kapture.records_camera)]
            logger.info(f'importing {len(filename_list)} image files ...')
            import_record_data_from_dir_auto(training_images_path,
                                             kapture_training_path,
                                             filename_list,
                                             images_import_method)
        # Import query images
        kapture_query_path = path.join(top_kaptures_path, f'slice{slice_n}', "query")
        delete_existing_kapture_files(kapture_query_path, force_erase=force_overwrite_existing)
        query_records_camera, query_trajectories = import_extended_cmu_seasons_images(query_image_list)
        query_kapture = kapture.Kapture(sensors=cameras,
                                        records_camera=query_records_camera,
                                        trajectories=query_trajectories)

        # import query gt when possible
        query_gt_kapture = []
        for query_gt_path in query_gt_list:
            query_gt_records_camera, query_gt_trajectories = import_extended_cmu_seasons_images(query_gt_path)
            query_gt_kapture.append(kapture.Kapture(sensors=cameras,
                                                    records_camera=query_gt_records_camera,
                                                    trajectories=query_gt_trajectories))
        data_to_merge = [query_kapture] + query_gt_kapture
        query_kapture = merge_keep_ids(data_to_merge, skip_list=[], data_paths=["" for _ in range(len(data_to_merge))],
                                       kapture_path="", images_import_method=TransferAction.skip)
        if import_all_files:
            _add_images_from_folder(query_images_path, query_kapture)
        kapture_to_dir(kapture_query_path, query_kapture)
        # finally import images
        if images_import_method != TransferAction.skip:
            filename_list = [f for _, _, f in kapture.flatten(query_kapture.records_camera)]
            logger.info(f'importing {len(filename_list)} image files ...')
            import_record_data_from_dir_auto(query_images_path, kapture_query_path, filename_list, images_import_method)


def import_extended_cmu_seasons_command_line() -> None:
    """
    Do the CMU dataset import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='imports Extended CMU Seasons dataset to the kapture format.'
                    'This script will create one kapture dataset'
                    ' for each of the 24 slices, based on the txt files')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to Extended CMU root folder')
    parser.add_argument('-o', '--output', required=True, help='output directory')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.skip,
                        help=f'How to import images [skip], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('--slice-range', nargs='+', default=list(range(2, 26)), type=int,
                        help='list of slices to import')
    parser.add_argument('--all-files', action='store_true', default=False,
                        help='Add images present in the folder yet not listed in the txt files.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_extended_cmu_seasons(args.input, args.output,
                                args.slice_range, args.all_files,
                                args.force, args.image_transfer)


if __name__ == '__main__':
    import_extended_cmu_seasons_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Convert kapture data in version 1.0 to version 1.1
"""

import logging
import os
import os.path as path
import argparse
from typing import Optional
import shutil
# import numpy as np like in kapture.io.csv
# so that types written as "np.float32" are understood by read_old_image_features_csv
import numpy as np  # noqa: F401

import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
import kapture.io.features
import kapture.io.csv
from kapture.io.records import import_record_data_from_dir_auto
from kapture.utils.paths import populate_files_in_dirpath
from kapture.io.binary import TransferAction
from kapture.utils.upgrade import CSV_FILENAMES_1_0, read_old_image_features_csv

logger = logging.getLogger('upgrade_1_0_to_1_1')


def upgrade_1_0_to_1_1(kapture_dirpath: str,  # noqa: C901: function a bit long but well documented
                       output_path: str,
                       keypoints_type: Optional[str],
                       descriptors_type: Optional[str],
                       global_features_type: Optional[str],
                       descriptors_metric_type: str,
                       global_features_metric_type: str,
                       images_import_strategy: TransferAction,
                       force_overwrite_existing: bool) -> None:
    """
    Convert kapture data in version 1.0 to version 1.1
    """
    os.makedirs(output_path, exist_ok=True)
    delete_existing_kapture_files(output_path, force_erase=force_overwrite_existing)
    # some text files didn't change, just change their header
    os.makedirs(path.join(output_path, 'sensors'), exist_ok=True)
    os.makedirs(path.join(output_path, 'reconstruction'), exist_ok=True)
    for csv_filename in CSV_FILENAMES_1_0:
        csv_fullpath = path.join(kapture_dirpath, csv_filename)
        if path.isfile(csv_fullpath):
            logger.debug(f'converting {csv_fullpath}...')
            old_version = kapture.io.csv.get_version_from_csv_file(csv_fullpath)
            if 'points3d' not in csv_filename:
                assert old_version == '1.0'
            else:
                assert (old_version is None or old_version == '1.0')
            csv_output_path = path.join(output_path, csv_filename)
            with open(csv_fullpath, 'r') as source_file:
                with open(csv_output_path, 'w') as output_file:
                    if old_version is not None:
                        source_file.readline()  # read and ignore header
                    # write replacement header
                    output_file.write(kapture.io.csv.KAPTURE_FORMAT_1 + kapture.io.csv.kapture_linesep)
                    shutil.copyfileobj(source_file, output_file)

    # keypoints
    keypoints_dir_path = path.join(kapture_dirpath, 'reconstruction', 'keypoints')
    keypoints_csv_path = path.join(keypoints_dir_path, 'keypoints.txt')
    if path.isdir(keypoints_dir_path) and path.isfile(keypoints_csv_path):
        logger.debug(f'converting {keypoints_dir_path}...')
        old_version = kapture.io.csv.get_version_from_csv_file(keypoints_csv_path)
        assert old_version == '1.0'
        name, dtype, dsize = read_old_image_features_csv(keypoints_csv_path)
        if keypoints_type is None:
            assert name != ''
            keypoints_type = name
        keypoints = kapture.Keypoints(name, dtype, dsize)
        keypoints_csv_output_path = path.join(output_path,
                                              kapture.io.csv.FEATURES_CSV_FILENAMES[kapture.Keypoints](keypoints_type))
        keypoints_output_dir = path.dirname(keypoints_csv_output_path)
        kapture.io.csv.keypoints_to_file(keypoints_csv_output_path, keypoints)
        # now copy all .kpt files
        keypoints_filenames = populate_files_in_dirpath(keypoints_dir_path, '.kpt')
        for keypoints_filename in keypoints_filenames:
            keypoints_output_file = path.join(keypoints_output_dir, keypoints_filename)
            os.makedirs(path.dirname(keypoints_output_file), exist_ok=True)
            shutil.copy(path.join(keypoints_dir_path, keypoints_filename), keypoints_output_file)

    # descriptors
    descriptors_dir_path = path.join(kapture_dirpath, 'reconstruction', 'descriptors')
    descriptors_csv_path = path.join(descriptors_dir_path, 'descriptors.txt')
    if path.isdir(descriptors_dir_path) and path.isfile(descriptors_csv_path):
        logger.debug(f'converting {descriptors_dir_path}...')
        old_version = kapture.io.csv.get_version_from_csv_file(descriptors_csv_path)
        assert old_version == '1.0'
        assert keypoints_type is not None
        name, dtype, dsize = read_old_image_features_csv(descriptors_csv_path)
        if descriptors_type is None:
            assert name != ''
            descriptors_type = name
        descriptors = kapture.Descriptors(name, dtype, dsize, keypoints_type, descriptors_metric_type)
        descriptors_csv_output_path = path.join(output_path,
                                                kapture.io.csv.FEATURES_CSV_FILENAMES[kapture.Descriptors](
                                                    descriptors_type)
                                                )
        descriptors_output_dir = path.dirname(descriptors_csv_output_path)
        kapture.io.csv.descriptors_to_file(descriptors_csv_output_path, descriptors)
        # now copy all .desc files
        descriptors_filenames = populate_files_in_dirpath(descriptors_dir_path, '.desc')
        for descriptors_filename in descriptors_filenames:
            descriptors_output_file = path.join(descriptors_output_dir, descriptors_filename)
            os.makedirs(path.dirname(descriptors_output_file), exist_ok=True)
            shutil.copy(path.join(descriptors_dir_path, descriptors_filename), descriptors_output_file)

    # matches
    matches_dir_path = path.join(kapture_dirpath, 'reconstruction', 'matches')
    if path.isdir(matches_dir_path):
        logger.debug(f'converting {matches_dir_path}...')
        assert keypoints_type is not None
        matches_output_dir = kapture.io.features.get_matches_fullpath(None, keypoints_type, output_path)
        # now copy all .matches files
        matches_filenames = populate_files_in_dirpath(matches_dir_path, '.matches')
        for matches_filename in matches_filenames:
            matches_output_file = path.join(matches_output_dir, matches_filename)
            os.makedirs(path.dirname(matches_output_file), exist_ok=True)
            shutil.copy(path.join(matches_dir_path, matches_filename), matches_output_file)

    # global features
    global_features_dir_path = path.join(kapture_dirpath, 'reconstruction', 'global_features')
    global_features_csv_path = path.join(global_features_dir_path, 'global_features.txt')
    if path.isdir(global_features_dir_path) and path.isfile(global_features_csv_path):
        logger.debug(f'converting {global_features_dir_path}...')
        old_version = kapture.io.csv.get_version_from_csv_file(global_features_csv_path)
        assert old_version == '1.0'
        assert keypoints_type is not None
        name, dtype, dsize = read_old_image_features_csv(global_features_csv_path)
        if global_features_type is None:
            assert name != ''
            global_features_type = name
        global_features = kapture.GlobalFeatures(name, dtype, dsize, global_features_metric_type)
        global_features_csv_output_path = path.join(output_path,
                                                    kapture.io.csv.FEATURES_CSV_FILENAMES[kapture.GlobalFeatures](
                                                        global_features_type)
                                                    )
        global_features_output_dir = path.dirname(global_features_csv_output_path)
        kapture.io.csv.global_features_to_file(global_features_csv_output_path, global_features)
        # now copy all .gfeat files
        global_features_filenames = populate_files_in_dirpath(global_features_dir_path, '.gfeat')
        for global_features_filename in global_features_filenames:
            global_features_output_file = path.join(global_features_output_dir, global_features_filename)
            os.makedirs(path.dirname(global_features_output_file), exist_ok=True)
            shutil.copy(path.join(global_features_dir_path, global_features_filename), global_features_output_file)

    # observations
    observations_csv_filename = path.join('reconstruction', 'observations.txt')
    observations_csv_path = path.join(kapture_dirpath, observations_csv_filename)
    if path.isfile(observations_csv_path):
        logger.debug(f'converting {observations_csv_path}...')
        old_version = kapture.io.csv.get_version_from_csv_file(observations_csv_path)
        assert old_version == '1.0'
        assert keypoints_type is not None
        observations = kapture.Observations()
        with open(observations_csv_path, 'r') as source_file:
            table = kapture.io.csv.table_from_file(source_file)
            # point3d_id, [image_path, feature_id]*
            for points3d_id_str, *pairs in table:
                points3d_id = int(points3d_id_str)
                if len(pairs) > 1:
                    image_paths = pairs[0::2]
                    keypoints_ids = pairs[1::2]
                    for image_path, keypoint_id in zip(image_paths, keypoints_ids):
                        observations.add(points3d_id, keypoints_type, image_path, int(keypoint_id))
        observations_output_path = path.join(output_path, observations_csv_filename)
        kapture.io.csv.observations_to_file(observations_output_path, observations)

    # records_data
    records_data_path = path.join(kapture_dirpath, 'sensors', 'records_data')
    logger.debug(f'converting {records_data_path}...')
    filename_list = list(populate_files_in_dirpath(records_data_path))
    import_record_data_from_dir_auto(records_data_path, output_path, filename_list, images_import_strategy)
    logger.debug('all done!')


def upgrade_1_0_to_1_1_command_line() -> None:
    """
    Convert kapture data in version 1.0 to version 1.1.
    """
    parser = argparse.ArgumentParser(
        description='convert kapture data in version 1.0 to version 1.1')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # export ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to kapture directory')
    parser.add_argument('-o', '--output', required=True, help='output directory.')
    parser.add_argument('--keypoints-type', default=None,
                        help='types of keypoints.')
    parser.add_argument('--descriptors-type', default=None, help='types of descriptors.')
    parser.add_argument('--descriptors-metric-type', default='L2', help='types of descriptors.')
    parser.add_argument('--global-features-type', default=None,
                        help='types of global features.')
    parser.add_argument('--global-features-metric-type', default='L2', help='types of descriptors.')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.skip,
                        help=f'How to import images [skip], '
                        f'choose among: {", ".join(a.name for a in TransferAction)}')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    upgrade_1_0_to_1_1(args.input, args.output,
                       args.keypoints_type, args.descriptors_type, args.global_features_type,
                       args.descriptors_metric_type, args.global_features_metric_type,
                       args.image_transfer,
                       args.force)


if __name__ == '__main__':
    upgrade_1_0_to_1_1_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Extract GPS from EXIF
"""
from typing import Optional
from PIL import Image
import piexif
import logging
from typing import Tuple, Dict
from tqdm import tqdm
import kapture
from kapture.io.csv import get_csv_fullpath
from kapture.io.records import images_to_filepaths
from warnings import warn


logger = logging.getLogger('exif')


def replace_exif_id_by_names(exif_dict: dict,
                             tag_dict: dict):
    """
    substitute TAG_ID with tag names when tag is referenced in tag_dict.

    :param exif_dict:
    :param tag_dict:
    :return:
    """
    return {
        tag_dict.get(key_id, key_id): value
        for key_id, value in exif_dict.items()
        if key_id in tag_dict
    }


def read_exif(
        image_filepath: str
) -> Dict:
    """
    Reads the EXIF metatdata from image file using piexif.

    :param image_filepath:
    :return:
    """
    with Image.open(image_filepath) as image:
        if 'exif' not in image.info:
            return None
        exif_data = piexif.load(image.info['exif'])
    return exif_data


def convert_rational_to_float(
        rational_num: Tuple[float]
) -> float:
    assert isinstance(rational_num, tuple)
    return float(rational_num[0] / rational_num[1])


def convert_dms_to_float(
        dms: Tuple[Tuple[float]],
        reference: Optional[str] = None
) -> float:
    """
    Convert the degree, minute, seconds + ref representation to single float.

    :param dms: Is a triplet (tuple) of IFDRational.
    :param reference:
    :return:
    """
    # sanity check
    assert isinstance(dms, tuple)
    assert len(dms) == 3
    if any(not isinstance(v, tuple) for v in dms):
        # Pillow 3.1 uses IFDRational, but piexif tuple[float].
        warn('Make sure you have installed piexif==1.1.3')
        raise TypeError('Expect Tuple[Tuplep[float] for sexagesimal. Make sure you have installed piexif==1.1.3')

    assert all(isinstance(v, tuple) for v in dms)
    decimal = convert_rational_to_float(dms[0]) + \
        convert_rational_to_float(dms[1]) / 60 + \
        convert_rational_to_float(dms[2]) / 3600
    # handle South or west coordinates.
    if reference is not None and reference.upper() in ['S', 'W']:
        decimal *= -1
    return decimal


def convert_gps_to_kapture_record(
        exif_data: dict
) -> kapture.RecordGnss:
    """
    Converts exif_data to kapture single RecordGnss.

    :param exif_data: input exif_data in EXIF format with tag nam (as given by read_exif).
        exif_data['GPS'] = {
            'piexif.GPSIFD.GPSLatitudeRef': 'N',
            'piexif.GPSIFD.GPSLatitude': ((52, 1), (31, 1), (810, 100)),
            'piexif.GPSIFD.GPSLongitudeRef': 'E',
            'piexif.GPSIFD.GPSLongitude': ((13, 1), (24, 1), (106, 100)),
            'piexif.GPSIFD.GPSAltitude': (27, 1),
            'piexif.GPSIFD.GPSDOP': (5, 1)
        }

    :return:
    """

    if 'GPS' not in exif_data:
        return kapture.RecordGnss()

    gps_info = exif_data['GPS']
    position = dict()
    position['x'] = convert_dms_to_float(
        dms=gps_info[piexif.GPSIFD.GPSLongitude],
        reference=gps_info.get(piexif.GPSIFD.GPSLongitudeRef)
    )
    position['y'] = convert_dms_to_float(
        dms=gps_info[piexif.GPSIFD.GPSLatitude],
        reference=gps_info.get(piexif.GPSIFD.GPSLatitudeRef)
    )
    position['z'] = convert_rational_to_float(
        rational_num=gps_info.get(piexif.GPSIFD.GPSAltitude, 0.0))
    position['dop'] = convert_rational_to_float(
        rational_num=gps_info.get(piexif.GPSIFD.GPSDOP, 0.0))
    position['utc'] = 0.

    return kapture.RecordGnss(**position)


def extract_gps_from_exif(
        kapture_data: kapture.Kapture,
        kapture_dirpath: str
):
    """
    Extract GPS coordinates from kapture dataset, returns the new sensor and gnss records.
    Gnss timestamps and sensor ids are guessed from timestamps and camera_id from images.
    The GNSS sensor_id are built prefixing 'GPS_'<cam_id>, with cam_id the sensor_id of the corresponding camera.

    :param kapture_data: input kapture data, must contains sensors and records_camera.
    :param kapture_dirpath: input path to kapture directory.
    :return:
    """
    # only load sensors + records_data:
    disable_tqdm = logger.getEffectiveLevel() != logging.INFO

    # make up new gps ids
    cam_to_gps_id = {  # cam_id -> gps_id
        cam_id: 'GPS_' + cam_id
        for cam_id, sensor in kapture_data.sensors.items()
        if sensor.sensor_type == kapture.SENSOR_TYPE_CAMERA
    }  # cam_id -> gps_id

    # set all gps to EPSG:4326
    gps_epsg_codes = {gps_id: 'EPSG:4326' for gps_id in cam_to_gps_id.values()}
    # add new gps ids to sensors
    gnss_kapture_sensors = kapture.Sensors()
    for gps_id, epsg in gps_epsg_codes.items():
        gnss_kapture_sensors[gps_id] = kapture.Sensor(sensor_type='gnss', sensor_params=[epsg])

    image_filepaths = images_to_filepaths(kapture_data.records_camera, kapture_dirpath=kapture_dirpath)
    records_gnss = kapture.RecordsGnss()

    for timestamp, cam_id, image_name in tqdm(kapture.flatten(kapture_data.records_camera), disable=disable_tqdm):
        image_filepath = image_filepaths[image_name]
        logger.debug(f'extracting GPS tags from {image_filepath}')
        gps_id = cam_to_gps_id[cam_id]
        exif_data = read_exif(image_filepath)
        gps_record = convert_gps_to_kapture_record(exif_data)
        records_gnss[timestamp, gps_id] = gps_record

    return gnss_kapture_sensors, records_gnss


def import_gps_from_exif(
        kapture_dirpath: str
):
    """
    Imports (extracts and writes) GPS data from EXIF metadata inside a kapture directory

    :param kapture_dirpath: path to kapture directory. kpautre data are modified.
    :return:
    """
    logger.info('loading kapture partial ...')
    skip_heavy_useless = [kapture.Trajectories,
                          kapture.RecordsLidar, kapture.RecordsWifi,
                          kapture.Keypoints, kapture.Descriptors, kapture.GlobalFeatures,
                          kapture.Matches, kapture.Points3d, kapture.Observations]
    kapture_data = kapture.io.csv.kapture_from_dir(kapture_dirpath, skip_list=skip_heavy_useless)

    # load exifs
    gnss_kapture_sensors, records_gnss = extract_gps_from_exif(kapture_data, kapture_dirpath)
    kapture_data.sensors.update(gnss_kapture_sensors)

    # overwrite sensors and gnss only
    sensors_filepath = get_csv_fullpath(kapture.Sensors, kapture_dirpath)
    logger.info(f'writing {sensors_filepath} ...')
    kapture.io.csv.sensors_to_file(sensors_filepath, kapture_data.sensors)
    records_gnss_filepath = get_csv_fullpath(kapture.RecordsGnss, kapture_dirpath)
    logger.info(f'writing {records_gnss_filepath} ...')
    kapture.io.csv.records_gnss_to_file(records_gnss_filepath, records_gnss)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Virtual gallery training data import.
"""

from collections import OrderedDict
import logging
import quaternion
from typing import Iterable, List
# kapture
import kapture
from kapture.algo.compare import equal_poses

# local
from .virtual_gallery_constants import virtual_gallery_camera_model, virtual_gallery_width, virtual_gallery_height


logger = logging.getLogger('virtual_gallery')


class VirtualGalleryTrainingIntrinsic:
    """
    virtual gallery training intrinsics
    """

    def __init__(self, camera_id: int, intrinsics: list):
        self.camera_id = camera_id
        self.intrinsics = intrinsics

    def __hash__(self):
        return hash((self.camera_id, tuple(self.intrinsics)))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.camera_id == other.camera_id and self.intrinsics == other.intrinsics


class VirtualGalleryTrainingExtrinsic:
    """
    virtual gallery training extrinsics
    """

    def __init__(self, frame_id: int, light_id: int, loop_id: int, camera_id: int, extrinsics: list):
        self.light_id = light_id
        self.loop_id = loop_id
        self.camera_id = camera_id
        self.extrinsics = extrinsics
        self.frame_id = frame_id

    def __hash__(self):
        return hash((self.light_id, self.loop_id, self.camera_id, tuple(self.extrinsics), self.frame_id))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.light_id == other.light_id and
                self.loop_id == other.loop_id and
                self.camera_id == other.camera_id and
                self.extrinsics == other.extrinsics and
                self.frame_id == other.frame_id)


def import_training_intrinsics(input_root: str,
                               light_range: list,
                               loop_range: list,
                               camera_range: set) -> Iterable[VirtualGalleryTrainingIntrinsic]:
    """
    Import training intrinsics

    # format of intrinsic.txt is
    # frame cameraID K[0,0] K[1,1] K[0,2] K[1,2]
    # 0 0 1371.022 1371.022 959.5 539.5

    :param input_root:
    :param light_range:
    :param loop_range:
    :param camera_range:
    :return: intrinsics
    """
    # assert that all intrinsics are identical for one camera and take any
    intrinsics = []
    for light_id in light_range:
        for loop_id in loop_range:
            with open(f'{input_root}/training/gallery_light{light_id}'
                      f'_loop{loop_id}/intrinsic.txt', 'r') as file:
                lines = file.readlines()
                for split in map(lambda x: x.split(), lines):
                    if split[0].isdigit() and int(split[1]) in camera_range:
                        intrinsic = VirtualGalleryTrainingIntrinsic(
                            int(split[1]),
                            [float(split[2]), float(split[3]), float(split[4]), float(split[5])]
                        )
                        intrinsics.append(intrinsic)
    return OrderedDict.fromkeys(intrinsics).keys()


def import_training_extrinsics(input_root: str,
                               light_range: list,
                               loop_range: list,
                               camera_range: set) -> List[VirtualGalleryTrainingExtrinsic]:
    """
    Import training extrinsics.

    # format of extrinsic.txt is
    # frame cameraID r1,1 r1,2 r1,3 t1 r2,1 r2,2 r2,3 t2 r3,1 r3,2 r3,3 t3 0 0
    # 0 1
    # 0 0 0.929348 0 0.3692049 2.798107 0 1 0 1.65 -0.3692049 0 0.929348
    # 1.618328 0 0 0 1

    :param input_root:
    :param light_range:
    :param loop_range:
    :param camera_range:
    :return: extrinsics
    """
    extrinsics = []
    for light_id in light_range:
        for loop_id in loop_range:
            with open(f'{input_root}/training/gallery_light{light_id}'
                      f'_loop{loop_id}/extrinsic.txt', 'r') as file:
                lines = file.readlines()
                for split in map(lambda x: x.split(), lines):
                    if split[0].isdigit() and int(split[1]) in camera_range:
                        extrinsic = VirtualGalleryTrainingExtrinsic(
                            int(split[0]), light_id, loop_id, int(split[1]),
                            [float(value) for value in split[2:]]
                        )
                        extrinsics.append(extrinsic)
    return extrinsics


def _get_training_camera_name(camera_id: int) -> str:
    return f'training_camera_{camera_id}'


def _get_training_rig_name() -> str:
    return 'training_rig'


# virtual gallery rig configuration
# camera are arranged in a circle (0° is cam0 forward then 60° each) and offset by 10cm
training_rig_config = kapture.Rigs({_get_training_rig_name():
                                    {_get_training_camera_name(0):
                                     kapture.PoseTransform(r=quaternion.quaternion(1, 0.0, 0.0, 0.0),
                                                           t=[0.0, 0.0, -0.10]),
                                     _get_training_camera_name(1):
                                     kapture.PoseTransform(r=quaternion.quaternion(-0.8660254037844386, 0.0, 0.5, 0.0),
                                                           t=[0.0, 0.0, -0.10]),
                                     _get_training_camera_name(2):
                                     kapture.PoseTransform(r=quaternion.quaternion(0.5, 0.0, -0.8660254037844386, 0.0),
                                                           t=[0.0, 0.0, -0.10]),
                                     _get_training_camera_name(3):
                                     kapture.PoseTransform(r=quaternion.quaternion(0.0, 0.0, -1, 0.0),
                                                           t=[0.0, 0.0, -0.10]),
                                     _get_training_camera_name(4):
                                     kapture.PoseTransform(r=quaternion.quaternion(-0.5, 0.0, -0.8660254037844386, 0.0),
                                                           t=[0.0, 0.0, -0.10]),
                                     _get_training_camera_name(5):
                                     kapture.PoseTransform(r=quaternion.quaternion(0.8660254037844386, 0.0, 0.5, 0.0),
                                                           t=[0.0, 0.0, -0.10])}})


def convert_training_intrinsics(training_intrinsics: Iterable[VirtualGalleryTrainingIntrinsic],
                                sensors: kapture.Sensors) -> None:
    """
    Import all training intrinsics into the sensors definitions.

    :param training_intrinsics: training intrinsics to import
    :param sensors: list of sensor definitions where to add the new definitions
    """
    logger.info("Converting training cameras...")
    for intrinsic in training_intrinsics:
        camera_device_id = _get_training_camera_name(intrinsic.camera_id)
        camera = kapture.Camera(virtual_gallery_camera_model,
                                [virtual_gallery_width, virtual_gallery_height] + intrinsic.intrinsics)
        sensors[camera_device_id] = camera


def convert_training_extrinsics(offset: int, training_extrinsics: Iterable[VirtualGalleryTrainingExtrinsic],
                                images: kapture.RecordsCamera, trajectories: kapture.Trajectories,
                                as_rig: bool) -> None:
    """
    Import all training extrinsics into the images and trajectories.

    :param offset:
    :param training_extrinsics: training extrinsics to import
    :param images: image list to add to
    :param trajectories: trajectories to add to
    :param as_rig: writes the position of the rig instead of individual cameras
    """
    # Map (light_id, loop_id, frame_id) to a unique timestamp
    training_frames_tuples = ((extrinsic.light_id, extrinsic.loop_id, extrinsic.frame_id)
                              for extrinsic in training_extrinsics)
    training_frames_tuples = OrderedDict.fromkeys(training_frames_tuples).keys()
    training_frame_mapping = {v: n + offset for n, v in enumerate(training_frames_tuples)}

    # Export images and trajectories
    logger.info("Converting training images and trajectories...")
    for extrinsic in training_extrinsics:
        rotation_matrix = [extrinsic.extrinsics[0:3], extrinsic.extrinsics[4:7], extrinsic.extrinsics[8:11]]
        rotation = quaternion.from_rotation_matrix(rotation_matrix)
        timestamp = training_frame_mapping[(extrinsic.light_id, extrinsic.loop_id, extrinsic.frame_id)]
        camera_device_id = _get_training_camera_name(extrinsic.camera_id)
        translation_vector = [extrinsic.extrinsics[3], extrinsic.extrinsics[7], extrinsic.extrinsics[11]]
        images[(timestamp, camera_device_id)] = (f'training/gallery_light{extrinsic.light_id}_loop'
                                                 f'{extrinsic.loop_id}/frames/rgb/camera_{extrinsic.camera_id}'
                                                 f'/rgb_{extrinsic.frame_id:05}.jpg')
        pose_cam = kapture.PoseTransform(rotation, translation_vector)
        if as_rig:
            pose_rig = kapture.PoseTransform.compose(
                [training_rig_config[(_get_training_rig_name(), camera_device_id)].inverse(),
                 pose_cam])
            if (timestamp, _get_training_rig_name()) in trajectories:
                assert equal_poses(pose_rig, trajectories[(timestamp, _get_training_rig_name())])
            else:
                trajectories[(timestamp,  _get_training_rig_name())] = pose_rig
        else:
            trajectories[(timestamp, camera_device_id)] = pose_cam



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to plot a kapture to a PLY file.
"""

import argparse
import logging
import os
import os.path as path
from tqdm import tqdm
import numpy as np
from PIL import Image
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
import kapture.io.csv as csv
import kapture.io.ply as ply
import kapture.io.image as image
import kapture.io.features
from kapture.io.records import records_depth_from_file
from typing import Optional

logger = logging.getLogger('plot')


def plot_ply(kapture_path: str,  # noqa: C901
             ply_path: str,
             keypoints_type: Optional[str],
             axis_length: float,
             only: Optional[list] = None,
             skip: Optional[list] = None
             ) -> None:
    """
    Plot the kapture data in a PLY file.

    :param kapture_path: top directory of the kapture
    :param ply_path: path to the ply file to create
    :param axis_length: length of axis representation (in world unit)
    :param only: list of the only kapture objects to plot (optional)
    :param skip: list of the kapture objects to skip
    """
    try:
        os.makedirs(ply_path, exist_ok=True)

        logger.info('loading data ...')
        with csv.get_all_tar_handlers(kapture_path) as tar_handlers:
            kapture_data = csv.kapture_from_dir(kapture_path, tar_handlers=tar_handlers)

            def _should_do(choice: str) -> bool:
                ok_only = only is None or choice in only
                ok_skip = skip is None or choice not in skip
                return ok_only or ok_skip

            logger.info('plotting  ...')
            if _should_do('rigs') and kapture_data.rigs:
                logger.info(f'creating {len(kapture_data.rigs)} rigs.')
                for rig_id, rig in kapture_data.rigs.items():
                    rig_ply_filepath = path.join(ply_path, f'rig_{rig_id}.ply')
                    logger.info(f'creating rig file : {rig_ply_filepath}.')
                    logger.debug(rig_ply_filepath)
                    ply.rig_to_ply(rig_ply_filepath, rig, axis_length)

            if _should_do('trajectories') and kapture_data.trajectories:
                trajectories_ply_filepath = path.join(ply_path, 'trajectories.ply')
                logger.info(f'creating trajectories file : {trajectories_ply_filepath}')
                ply.trajectories_to_ply(trajectories_ply_filepath, kapture_data.trajectories, axis_length)

            if _should_do('points3d') and kapture_data.points3d:
                points3d_ply_filepath = path.join(ply_path, 'points3d.ply')
                logger.info(f'creating 3D points file : {points3d_ply_filepath}')
                ply.points3d_to_ply(points3d_ply_filepath, kapture_data.points3d)

            if _should_do('keypoints') and kapture_data.keypoints:
                if keypoints_type is None:
                    if len(kapture_data.keypoints) == 1:
                        keypoints_type = next(iter(kapture_data.keypoints.keys()))
                    else:
                        raise ValueError('should_do(keypoints) is true but keypoints_type '
                                         'is None and could be multiple values, please specify')

                logger.info('creating keypoints in 3D : '
                            f'{kapture.io.features.get_keypoints_fullpath(keypoints_type, ply_path)}')
                keypoints_dsize = kapture_data.keypoints[keypoints_type].dsize
                keypoints_dtype = kapture_data.keypoints[keypoints_type].dtype
                keypoints_filepaths = kapture.io.features.keypoints_to_filepaths(kapture_data.keypoints[keypoints_type],
                                                                                 keypoints_type,
                                                                                 kapture_path,
                                                                                 tar_handlers)
                for image_filename, keypoints_filepath in tqdm(keypoints_filepaths.items(),
                                                               disable=logger.level >= logging.CRITICAL):
                    image_filepath = kapture.io.records.get_image_fullpath(kapture_path, image_filename)
                    image_keypoints_filepath = kapture.io.features.get_keypoints_fullpath(keypoints_type,
                                                                                          ply_path,
                                                                                          image_filename) + '.jpg'
                    image.image_keypoints_to_image_file(
                        image_keypoints_filepath, image_filepath, keypoints_filepath, keypoints_dtype, keypoints_dsize)

            if _should_do('depth') and kapture_data.records_depth:
                logger.info('creating depth maps in 3D.')
                depth_records_filepaths = kapture.io.records.depth_maps_to_filepaths(kapture_data.records_depth,
                                                                                     kapture_path)
                map_depth_to_sensor = {depth_map_name: sensor_id
                                       for _, sensor_id, depth_map_name in kapture.flatten(kapture_data.records_depth)}
                for depth_map_name, depth_map_filepath in tqdm(depth_records_filepaths.items(),
                                                               disable=logger.level >= logging.CRITICAL):
                    depth_png_filepath = path.join(ply_path, f'depth_images/{depth_map_name}.png')
                    logger.debug(f'creating depth map file {depth_png_filepath}')
                    os.makedirs(path.dirname(depth_png_filepath), exist_ok=True)
                    depth_sensor_id = map_depth_to_sensor[depth_map_name]
                    depth_map_sizes = tuple(int(x) for x in kapture_data.sensors[depth_sensor_id].sensor_params[1:3])
                    depth_map = records_depth_from_file(depth_map_filepath, depth_map_sizes)
                    # min max scaling
                    depth_map = (depth_map - depth_map.min()) / (depth_map.max() - depth_map.min())
                    depth_map = (depth_map * 255.).astype(np.uint8)
                    depth_image = Image.fromarray(depth_map)
                    depth_image.save(depth_png_filepath)
            logger.info('done.')
    except Exception as e:
        logging.critical(e)
        raise


def export_ply_command_line() -> None:
    """
    Do the plot to ply file using the parameters given on the command line.
    """
    plot_choices = {
        # cmd : help
        'rigs': 'plot the rig geometry, ie. relative pose of sensors into the rig.',
        'rig_stat': 'plot the sensor relative poses for each trajectory timestamp.',
        'trajectories': 'plot the trajectory of every sensors.',
        'points3d': 'plot the 3-D point cloud.',
        'keypoints': 'plot keypoints in 3D over the image plane.',
        'depth': 'plot depth maps as point cloud (one per depth map)'
    }

    parser = argparse.ArgumentParser(description='plot out camera geometry to PLY file.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-i', '--input', required=True,
                        help='input path to kapture data root directory')
    parser.add_argument('-o', '--output', required=False,
                        help='output directory (PLY file format).')
    parser.add_argument('--only', nargs='+', choices=plot_choices.keys(), default=[],
                        help='things to plot : ' + ' // '.join('{}: {}'.format(k, v) for k, v in plot_choices.items()))
    parser.add_argument('--keypoints-type', default=None, help='types of keypoints.')
    parser.add_argument('--skip', nargs='+', choices=plot_choices.keys(), default=['keypoints'],
                        help='things to not plot : ' + ' // '.join(plot_choices.keys()))
    parser.add_argument('--axis_length', type=float, default=0.1,
                        help='length of axis representation (in world unit).')
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)
    if not args.output:
        args.output = args.input

    # only overwrite skip :
    if args.only:
        args.skip = []
    logger.debug(''.join(['\n\t{:13} = {}'.format(k, v) for k, v in vars(args).items()]))
    plot_ply(args.input, args.output, args.keypoints_type, args.axis_length, args.only, args.skip)


if __name__ == '__main__':
    export_ply_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from typing import Dict, List, Tuple, Union


class Observations(Dict[int, Dict[str, List[Tuple[str, int]]]]):
    """
    Observations. This can be used like this:

    - observations[point3d_idx][keypoints_type] = list( observation )
    - observation = (image_path, keypoint_idx)
    """

    def add(self, point3d_idx: int, keypoints_type: str, image_filename: str, keypoint_idx: int):
        """
        Adds a 2-D observation (image, keypoint) of a 3D point.

        :param point3d_idx: index of the 3D point to add an observation of.
        :param keypoints_type: type of keypoints, name of the keypoints subfolder
        :param image_filename: name of the image where the 3D points is observed
        :param keypoint_idx: index of the keypoints in the image that correspond to the 3D point.
        :return:
        """
        # enforce type checking
        if not isinstance(point3d_idx, int):
            raise TypeError('invalid type for point3d_idx')
        if not isinstance(keypoints_type, str):
            raise TypeError('invalid type for keypoints_type')
        if not isinstance(image_filename, str):
            raise TypeError('invalid type for image_filename')
        if not isinstance(keypoint_idx, int):
            raise TypeError('invalid type for keypoint_idx')
        self.setdefault(point3d_idx, {}).setdefault(keypoints_type, []).append((image_filename, keypoint_idx))

    def __getitem__(self, key: Union[int, Tuple[int, str]]) -> Union[Dict[str, List[Tuple[str, int]]],
                                                                     List[Tuple[str, int]]]:
        if isinstance(key, tuple):
            # key is a pair of (point3d_idx, keypoints_type)
            point3d_idx = key[0]
            keypoints_type = key[1]
            if not isinstance(point3d_idx, int):
                raise TypeError('invalid point3d_idx')
            if not isinstance(keypoints_type, str):
                raise TypeError('invalid keypoints_type')
            return super(Observations, self).__getitem__(point3d_idx)[keypoints_type]
        elif isinstance(key, int):
            # key is a point3d_idx
            return super(Observations, self).__getitem__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def key_pairs(self) -> List[Tuple[int, str]]:
        """
        Returns the list of (point3d_idx, keypoints_type) contained in observations.
        Those pairs can be used to access a list of observation.
        :return: list of (point3d_idx, keypoints_type)
        """
        return [
            (point3d_idx, keypoints_type)
            for point3d_idx, per_feature_observations in self.items()
            for keypoints_type in per_feature_observations.keys()
        ]

    def __contains__(self, key: Union[int, Tuple[int, str]]):
        if isinstance(key, tuple):
            # key is a pair of (point3d_idx, keypoints_type)
            point3d_idx = key[0]
            keypoints_type = key[1]
            if not isinstance(point3d_idx, int):
                raise TypeError('invalid point3d_idx')
            if not isinstance(keypoints_type, str):
                raise TypeError('invalid keypoints_type')
            return super(Observations, self).__contains__(point3d_idx) and keypoints_type in self[point3d_idx]
        elif isinstance(key, int):
            return super(Observations, self).__contains__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __repr__(self) -> str:
        representation = ''
        # [point3d_idx, keypoints_type]:   (image_path, keypoint_idx)   (image_path, keypoint_idx)...
        for point3d_idx, keypoints_type in sorted(self.key_pairs(), key=lambda x: x[0]):
            representation += f'[{point3d_idx:05}, {keypoints_type}]: '
            assert point3d_idx is not None
            for image_path, keypoint_idx in self.get(point3d_idx)[keypoints_type]:
                representation += f'\t({image_path}, {keypoint_idx})'
            representation += '\n'
        return representation



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports an IDL dataset cvpr17 to the kapture format.


    A Dataset for Benchmarking Image-based Localization
    http://openaccess.thecvf.com/content_cvpr_2017/html/Sun_A_Dataset_for_CVPR_2017_paper.html

    Xun Sun, Yuanfan Xie, Pei Luo, Liang Wang
    Baidu Autonomous Driving Business Unit

    https://sites.google.com/site/xunsunhomepage/

    Data format:
    =================

    The camera poses are saved in the following format:

    fx 0 cx
    0 fy cy
    0 0 1
    0 0 0
    R'
    cop
    w h

"""

import os
import os.path as path
import logging
import argparse
from PIL import Image
import quaternion
from typing import Union

# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.utils.paths import path_secure


logger = logging.getLogger('IDL_dataset_cvpr17')


def import_idl_dataset_cvpr17(idl_dataset_path: str,
                              gt_path: Union[str, None],
                              kapture_path: str,
                              force_overwrite_existing: bool = False,
                              images_import_method: TransferAction = TransferAction.skip) -> None:
    """
    Reads the IDL dataset and copy it to a kapture.

    :param idl_dataset_path: path to the IDL dataset
    :param gt_path: ground truth data path
    :param kapture_path: path to the kapture top directory to create
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    """

    os.makedirs(kapture_path, exist_ok=True)
    delete_existing_kapture_files(kapture_path, force_erase=force_overwrite_existing)

    cameras = kapture.Sensors()
    images = kapture.RecordsCamera()
    trajectories = kapture.Trajectories()

    file_list = [os.path.relpath(os.path.join(dirpath, filename), idl_dataset_path)
                 for dirpath, dirs, filenames in os.walk(idl_dataset_path)
                 for filename in filenames]
    file_list = sorted(file_list)

    logger.info('starting conversion...')
    for n, filename in enumerate(file_list):
        # test if file is a valid image
        try:
            # lazy load
            with Image.open(path.join(idl_dataset_path, filename)) as im:
                width, height = im.size
                model_params = [width, height]
        except Exception:
            continue

        camera_id = f'sensor{n}'
        images[(n, camera_id)] = path_secure(filename)  # don't forget windows
        model = kapture.CameraType.UNKNOWN_CAMERA
        if gt_path is not None:
            # replace image extension with .camera
            file_gt_path = os.path.splitext(os.path.join(gt_path, filename))[0] + ".camera"

            if os.path.isfile(file_gt_path):
                with open(file_gt_path) as fin:
                    lines = fin.readlines()
                    lines = (line.rstrip().split() for line in lines)  # split fields
                    lines = list(lines)
                fx = float(lines[0][0])
                cx = float(lines[0][2])
                fy = float(lines[1][1])
                cy = float(lines[1][2])
                width_file = float(lines[8][0])
                height_file = float(lines[8][1])
                assert (width_file == width)
                assert (height_file == height)
                model = kapture.CameraType.PINHOLE
                model_params = [width, height, fx, fy, cx, cy]

                rotation_matrix = [[float(v) for v in line] for line in lines[4:7]]
                rotation = quaternion.from_rotation_matrix(rotation_matrix)
                center_of_projection = [float(v) for v in lines[7]]
                pose = kapture.PoseTransform(rotation, center_of_projection).inverse()
                trajectories[(n, camera_id)] = pose
        cameras[camera_id] = kapture.Camera(model, model_params)

    # if no trajectory were added, no need to create the file
    if not trajectories:
        trajectories = None

    # import (copy) image files.
    logger.info('import image files ...')
    filename_list = [f for _, _, f in kapture.flatten(images)]
    import_record_data_from_dir_auto(idl_dataset_path, kapture_path, filename_list, images_import_method)

    # pack into kapture format
    imported_kapture = kapture.Kapture(sensors=cameras, records_camera=images, trajectories=trajectories)
    logger.info('writing imported data...')
    kapture_to_dir(kapture_path, imported_kapture)


def import_idl_dataset_cvpr17_command_line() -> None:
    """
    Do the IDL dataset import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='import IDL_dataset_cvpr17 to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to images folder')
    parser.add_argument('-gt', '--gt_path', default=None, help='path to gt data')
    parser.add_argument('-o', '--output', required=True, help='output directory')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_idl_dataset_cvpr17(
        idl_dataset_path=args.input,
        gt_path=args.gt_path,
        kapture_path=args.output,
        force_overwrite_existing=args.force,
        images_import_method=args.image_transfer)


if __name__ == '__main__':
    import_idl_dataset_cvpr17_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
All kapture objects representing the data kapture manages.
"""

from .PoseTransform import PoseTransform  # noqa: F401
from .Sensors import Sensor, Sensors, create_sensor  # noqa: F401
from .Sensors import Camera, CameraType, CAMERA_TYPE_PARAMS_COUNT, CAMERA_TYPE_PARAMS_COUNT_FROM_NAME  # noqa: F401
from .Sensors import SENSOR_TYPE_CAMERA, SENSOR_TYPE_DEPTH_CAM, ALL_CAMERA_SENSOR_TYPES  # noqa: F401
from .Rigs import Rigs  # noqa: F401
from .Trajectories import Trajectories  # noqa: F401
from .Trajectories import rigs_remove, rigs_remove_inplace, rigs_recover, rigs_recover_inplace  # noqa: F401
from .Records import RecordsBase, RecordsArray, RecordsCamera, RecordsDepth, RecordsLidar  # noqa: F401
from .Records import RecordsWifi, RecordWifi, RecordWifiSignal  # noqa: F401
from .Records import RecordBluetooth, RecordsBluetooth, RecordBluetoothSignal  # noqa: F401
from .Records import RecordGnss, RecordsGnss  # noqa: F401
from .Records import RecordAccelerometer, RecordsAccelerometer  # noqa: F401
from .Records import RecordGyroscope, RecordsGyroscope  # noqa: F401
from .Records import RecordMagnetic, RecordsMagnetic  # noqa: F401
from .ImageFeatures import Keypoints, Descriptors, GlobalFeatures  # noqa: F401
from .Observations import Observations  # noqa: F401
from .Matches import Matches  # noqa: F401
from .Points3d import Points3d  # noqa: F401
from .Kapture import Kapture  # noqa: F401
from .flatten import flatten  # noqa: F401

import logging

# it is up to the library client to specify the log handler.
logging.getLogger("kapture").addHandler(logging.NullHandler())



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Functions to export the reconstruction part of colmap (only text format is supported).
"""

from kapture.io.tar import TarCollection
import logging
import os.path as path
from typing import Dict, Optional, Tuple, List

# kapture
import kapture
import kapture.io.features
# local
from .cameras import CAMERA_MODEL_NAMES, get_colmap_camera

logger = logging.getLogger('colmap')


def export_to_colmap_cameras_txt(colmap_cameras_filepath: str,
                                 sensors: kapture.Sensors,
                                 colmap_camera_ids: Dict[str, int]) -> None:
    """
    Exports to colmap reconstruction file "cameras.txt".

    :param colmap_cameras_filepath: path to colmap file "cameras.txt" to be writen.
    :param sensors: sensors to be exported
    :param colmap_camera_ids: gives the correspondences between kapture camera id and colmap camera id
    """
    assert path.basename(colmap_cameras_filepath) == 'cameras.txt'
    assert isinstance(colmap_camera_ids, dict)
    assert isinstance(sensors, kapture.Sensors)
    cameras = {cam_id: cam
               for cam_id, cam in sensors.items()
               if isinstance(cam, kapture.Camera)}

    # cameras.txt
    cameras_colmap_header = '# Sensor list with one line of data per camera:\n' \
                            '#   CAMERA_ID, MODEL, WIDTH, HEIGHT, PARAMS[]\n' \
                            '# NB cameras : {}\n'.format(len(cameras))

    converted_cameras = [(cam_id, get_colmap_camera(cam)) for cam_id, cam in cameras.items()]
    with open(colmap_cameras_filepath, 'w') as fid:
        fid.write(cameras_colmap_header)
        lines = (' '.join([str(colmap_camera_ids[cam_id]), CAMERA_MODEL_NAMES[col_cam_id]] +
                          [str(int(width)), str(int(height))] +
                          [str(int(v)) if v.is_integer() else str(v) for v in params])
                 for cam_id, (col_cam_id, width, height, params, prior_focal_length) in converted_cameras
                 if cam_id in colmap_camera_ids)
        fid.write('\n'.join(lines))


def export_to_colmap_images_txt(colmap_images_filepath: str,
                                images: kapture.RecordsCamera,
                                trajectories: kapture.Trajectories,
                                colmap_camera_ids: Dict[str, int],
                                colmap_image_ids: Dict[str, int],
                                image_to_keypoints: Dict[str, List[Tuple[float, float, int]]]) -> None:
    """
    Exports kapture to colmap reconstruction file "images.txt".

    :param colmap_images_filepath: path to colmap file "images.txt" to be writen.
    :param images: images list to export
    :param trajectories: poses to export
    :param colmap_camera_ids: correspondences between kapture camera id and colmap camera id
    :param colmap_image_ids: correspondences between kapture image id (image path) and colmap image id
    :param image_to_keypoints: input image_filename -> [(x, y, point_id), (x, y, point_id), ...]
    """
    assert path.basename(colmap_images_filepath) == 'images.txt'
    assert isinstance(images, kapture.RecordsCamera)
    assert isinstance(trajectories, kapture.Trajectories)
    assert isinstance(colmap_camera_ids, dict)
    assert isinstance(colmap_image_ids, dict)
    assert isinstance(image_to_keypoints, dict)

    images_flattened = list(kapture.flatten(images))
    images_colmap_header = '# Image list with two lines of data per image:\n' \
                           '#   IMAGE_ID, QW, QX, QY, QZ, TX, TY, TZ, CAMERA_ID, NAME\n' \
                           '#   POINTS2D[] as (X, Y, POINT3D_ID)\n' \
                           '# NB IMAGES : {}\n'.format(len(images_flattened))

    with open(colmap_images_filepath, 'w') as fid:
        fid.write(images_colmap_header)
        for timestamp, sensor_id, sensing_filepath in images_flattened:
            colmap_cam_id = colmap_camera_ids[sensor_id]
            colmap_image_id = colmap_image_ids[sensing_filepath]
            # retrieve image pose from trajectories

            if timestamp not in trajectories:
                logging.debug('timestamp:{} not in trajectories'.format(timestamp))
                continue

            if sensor_id not in trajectories[timestamp]:
                logger.debug('camera {} not found in trajectories for timestamp {}.'.format(
                    sensor_id, timestamp))
                continue

            pose_tr = trajectories[timestamp].get(sensor_id)
            line = [colmap_image_id] + pose_tr.r_raw + pose_tr.t_raw + [colmap_cam_id, sensing_filepath]
            fid.write(' '.join('{}'.format(i) for i in line) + '\n')
            if sensing_filepath in image_to_keypoints:
                # POINTS2D[] as (X, Y, POINT3D_ID)
                p2d = [(str(x), str(y), str(point_id))
                       for x, y, point_id in image_to_keypoints[sensing_filepath]]
                fid.write(' '.join('{} {} {}'.format(i, j, k) for i, j, k in p2d))
            fid.write('\n')


def export_to_colmap_matches_txt(colmap_matches_filepath: str, matches: kapture.Matches) -> None:
    """
    Exports to colmap reconstruction file "matches.txt". This files can be used to redo the geometric filtering.

    :param colmap_matches_filepath: input path to colmap matches file to be writen.
    :param matches: input kapture.Matches
    """
    assert isinstance(matches, kapture.Matches)
    with open(colmap_matches_filepath, 'w') as fid:
        # matches[(image_path1, image_path2)] = image_matches
        for image_path1, image_path2 in matches:
            fid.write('{} {}\n'.format(image_path1, image_path2))


def export_to_colmap_points3d_txt(colmap_points3d_filepath: str,
                                  colmap_image_ids: Dict[str, int],
                                  points3d: kapture.Points3d = None,
                                  observations: kapture.Observations = None,
                                  keypoints_type: str = None) -> None:
    """
    Exports to colmap points3d text file.

    :param colmap_points3d_filepath: path to colmap points3d file to be writen.
    :param colmap_image_ids: correspondences between kapture image id (image path) and colmap image id
    :param points3d: kapture points3d to export
    :param observations: kapture observations to export
    :param keypoints_type: type of keypoints for which the observations are exported, name of the keypoints subfolder
    """
    assert isinstance(points3d, kapture.Points3d) or points3d is None
    assert isinstance(observations, kapture.Observations) or observations is None

    assert isinstance(colmap_image_ids, dict)
    points3d_colmap_header = '# 3D point list with one line of data per point:\n' \
                             '#   POINT3D_ID, X, Y, Z, R, G, B, ERROR, TRACK[] as (IMAGE_ID, POINT2D_IDX)\n'
    with open(colmap_points3d_filepath, 'w') as fid:
        fid.write(points3d_colmap_header)
        if points3d:
            for i in range(points3d.shape[0]):
                point3d = points3d[i]
                line = '{} {} {} {} {} {} {} 0'.format(i,
                                                       point3d[0], point3d[1], point3d[2],
                                                       int(point3d[3]), int(point3d[4]), int(point3d[5]))
                if observations is not None and i in observations:
                    if len(observations[i]) == 0:
                        continue
                    elif len(observations[i]) == 1 and keypoints_type is None:
                        keypoints_type = next(iter(observations[i].keys()))
                    if keypoints_type is None or keypoints_type not in observations[i]:
                        continue
                    if len(observations[i, keypoints_type]) > 0:
                        line += ' '
                        pairs = [(str(colmap_image_ids[name]), str(keypoint_id))
                                 for name, keypoint_id in observations[i, keypoints_type]]
                        line += ' '.join([str(s) for s in list(sum(pairs, ()))])
                line += '\n'
                fid.write(line)


def export_to_colmap_txt(colmap_reconstruction_dirpath: str,
                         kapture_data: kapture.Kapture,
                         kapture_dirpath: str,
                         tar_handler: Optional[TarCollection],
                         colmap_camera_ids: Dict[str, int],
                         colmap_image_ids: Dict[str, int],
                         keypoints_type: str = None) -> None:
    """
    Exports to colmap reconstruction text files.

    :param colmap_reconstruction_dirpath: path to directory where colmap reconstruction files will be stored.
    :param kapture_data: input kapture data
    :param kapture_dirpath: path to output directory, where colmap files will be stored
    :param tar_handler: collection of preloaded tar archives
    :param colmap_camera_ids: gives the correspondences between kapture camera id and colmap camera id
    :param colmap_image_ids: gives the correspondences between kapture image id (image path) and colmap image id
    :param keypoints_type: reference the the keypoints for which the reconstruction are exported
    """
    assert isinstance(kapture_data, kapture.Kapture)
    assert isinstance(colmap_camera_ids, dict)
    assert isinstance(colmap_image_ids, dict)
    assert kapture_data.records_camera is not None
    assert kapture_data.sensors is not None

    if kapture_data.rigs:
        # check the rigs are not used in trajectories
        used_sensors = set(sensor_id
                           for shot in kapture_data.trajectories.values()
                           for sensor_id in shot.keys())
        rigs_ids = set(rig_id for rig_id in kapture_data.rigs.keys())
        if any(used_sensor in rigs_ids for used_sensor in used_sensors):
            raise ValueError(
                'colmap format does not handle rigs notation. '
                'Remove rig from trajectories beforehand (see rigs_remove_inplace)')

    # cameras.txt
    logger.info('creating colmap cameras.txt')
    export_to_colmap_cameras_txt(path.join(colmap_reconstruction_dirpath, 'cameras.txt'),
                                 kapture_data.sensors, colmap_camera_ids)

    # images.txt
    image_to_keypoints = {}
    if kapture_data.keypoints and kapture_data.points3d and kapture_data.observations:
        if len(kapture_data.keypoints) == 1 and keypoints_type is None:
            keypoints_type = next(iter(kapture_data.keypoints.keys()))
        if keypoints_type is not None and keypoints_type in kapture_data.keypoints:
            obs_for_keypoints_type = {point_id: per_keypoints_type_subdict[keypoints_type]
                                      for point_id, per_keypoints_type_subdict in kapture_data.observations.items()
                                      if keypoints_type in per_keypoints_type_subdict}
            observations_reversed = {(image_filename, keypoint_idx): point3d_idx
                                     for point3d_idx in obs_for_keypoints_type.keys()
                                     for image_filename, keypoint_idx in obs_for_keypoints_type[point3d_idx]}
            # prepare images.txt even lines
            # POINTS2D[] as (X, Y, POINT3D_ID)
            keypoints = kapture_data.keypoints[keypoints_type]
            keypoints_filepaths = kapture.io.features.keypoints_to_filepaths(keypoints,
                                                                             keypoints_type,
                                                                             kapture_dirpath,
                                                                             tar_handler)
            for image_filename, image_keypoints_filepath in keypoints_filepaths.items():
                image_keypoints = kapture.io.features.image_keypoints_from_file(image_keypoints_filepath,
                                                                                keypoints.dtype,
                                                                                keypoints.dsize)
                image_to_keypoints[image_filename] = []
                for i in range(image_keypoints.shape[0]):
                    point3d_idx = observations_reversed[(image_filename, i)] if (image_filename,
                                                                                 i) in observations_reversed else -1
                    image_to_keypoints[image_filename].append((image_keypoints[i, 0],
                                                               image_keypoints[i, 1],
                                                               point3d_idx))

    if kapture_data.records_camera is None or kapture_data.trajectories is None:
        logger.info('skipping colmap images.txt (missing images or trajectories).')
    else:
        logger.info('creating colmap images.txt')
        export_to_colmap_images_txt(path.join(colmap_reconstruction_dirpath, 'images.txt'),
                                    kapture_data.records_camera, kapture_data.trajectories,
                                    colmap_camera_ids,
                                    colmap_image_ids,
                                    image_to_keypoints)

    # image_matches.txt: to be imported as custom match (allow geometric verification)
    if kapture_data.matches:
        if len(kapture_data.matches) == 1 and keypoints_type is None:
            keypoints_type = next(iter(kapture_data.matches.keys()))
        if keypoints_type is not None and keypoints_type in kapture_data.matches:
            logger.info('creating image_matches.txt')
            export_to_colmap_matches_txt(path.join(colmap_reconstruction_dirpath, 'image_matches.txt'),
                                         kapture_data.matches[keypoints_type])

    # points3D.txt
    logging.info('creating colmap points3D.txt')
    export_to_colmap_points3d_txt(path.join(colmap_reconstruction_dirpath, 'points3D.txt'), colmap_image_ids,
                                  kapture_data.points3d, kapture_data.observations, keypoints_type)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from .Sensors import Sensors, Camera
from .Rigs import Rigs
from .Trajectories import Trajectories
from .Records import RecordsCamera, RecordsDepth, RecordsLidar, RecordsWifi, RecordsBluetooth
from .Records import RecordsGnss, RecordsAccelerometer, RecordsGyroscope, RecordsMagnetic
from .ImageFeatures import Keypoints, Descriptors, GlobalFeatures
from .Matches import Matches
from .Observations import Observations
from .Points3d import Points3d
from typing import Dict, Optional, Union
import logging

logger = logging.getLogger(__name__)


class Kapture:
    """
    brief: Root class of all kapture data.
    """

    def __init__(
            self,
            sensors: Optional[Sensors] = None,
            rigs: Optional[Rigs] = None,
            trajectories: Optional[Trajectories] = None,
            records_camera: Optional[RecordsCamera] = None,
            records_depth: Optional[RecordsDepth] = None,
            records_lidar: Optional[RecordsLidar] = None,
            records_wifi: Optional[RecordsWifi] = None,
            records_bluetooth: Optional[RecordsBluetooth] = None,
            records_gnss: Optional[RecordsGnss] = None,
            records_accelerometer: Optional[RecordsAccelerometer] = None,
            records_gyroscope: Optional[RecordsGyroscope] = None,
            records_magnetic: Optional[RecordsMagnetic] = None,
            keypoints: Optional[Dict[str, Keypoints]] = None,
            descriptors: Optional[Dict[str, Descriptors]] = None,
            global_features: Optional[Dict[str, GlobalFeatures]] = None,
            matches: Optional[Dict[str, Matches]] = None,
            observations: Optional[Observations] = None,
            points3d: Optional[Points3d] = None,
    ):
        self.sensors = sensors
        self.rigs = rigs
        self.trajectories = trajectories
        self.records_camera = records_camera
        self.records_depth = records_depth
        self.records_lidar = records_lidar
        self.records_wifi = records_wifi
        self.records_bluetooth = records_bluetooth
        self.records_gnss = records_gnss
        self.records_accelerometer = records_accelerometer
        self.records_gyroscope = records_gyroscope
        self.records_magnetic = records_magnetic
        self.keypoints = keypoints
        self.descriptors = descriptors
        self.global_features = global_features
        self.matches = matches
        self.observations = observations
        self.points3d = points3d
        self.__version__ = '0.0'

    @property
    def format_version(self):
        """
        :return: kapture format version
        """
        return self.__version__

    @property
    def sensors(self) -> Optional[Sensors]:
        """
        :return: the list of sensors
        """
        return self._sensors

    @sensors.setter
    def sensors(self, sensors: Optional[Sensors]):
        if sensors is not None and not isinstance(sensors, Sensors):
            raise TypeError('sensors expected')
        self._sensors = sensors

    @property
    def cameras(self) -> Dict[str, Camera]:
        """
        :return: the cameras (particular kind of sensor) as dictionary keyed by the camera identifier
        """
        if self.sensors is None:
            return {}
        return {k: v for k, v in self.sensors.items() if isinstance(v, Camera)}

    @property
    def rigs(self) -> Optional[Rigs]:
        """
        :return: the list of rigs
        """
        return self._rigs

    @rigs.setter
    def rigs(self, rigs: Optional[Rigs]):
        if rigs is not None and not isinstance(rigs, Rigs):
            raise TypeError('rigs expected')
        self._rigs = rigs

    @property
    def trajectories(self) -> Optional[Trajectories]:
        """
        :return: the list of trajectories
        """
        return self._trajectories

    @trajectories.setter
    def trajectories(self, trajectories: Optional[Trajectories]):
        if trajectories is not None and not isinstance(trajectories, Trajectories):
            raise TypeError('trajectories expected')
        self._trajectories = trajectories

    @property
    def records_camera(self) -> Optional[RecordsCamera]:
        """
        :return: Camera records
        """
        return self._records_camera

    @records_camera.setter
    def records_camera(self, records_camera: Optional[RecordsCamera]):
        if records_camera is not None and not isinstance(records_camera, RecordsCamera):
            raise TypeError('RecordsCamera expected')
        self._records_camera = records_camera

    @property
    def records_depth(self) -> Optional[RecordsDepth]:
        """
        :return: Depth records
        """
        return self._records_depth

    @records_depth.setter
    def records_depth(self, records_depth: Optional[RecordsDepth]):
        if records_depth is not None and not isinstance(records_depth, RecordsDepth):
            raise TypeError('RecordsDepth expected')
        self._records_depth = records_depth

    @property
    def records_lidar(self) -> Optional[RecordsLidar]:
        """
        :return: Lidar records
        """
        return self._records_lidar

    @records_lidar.setter
    def records_lidar(self, records_lidar: Optional[RecordsLidar]):
        if records_lidar is not None and not isinstance(records_lidar, RecordsLidar):
            raise TypeError('RecordsLidar expected')
        self._records_lidar = records_lidar

    @property
    def records_wifi(self) -> Optional[RecordsWifi]:
        """
        :return: Wifi records
        """
        return self._records_wifi

    @records_wifi.setter
    def records_wifi(self, records_wifi: Optional[RecordsWifi]):
        if records_wifi is not None and not isinstance(records_wifi, RecordsWifi):
            raise TypeError('RecordsWifi expected')
        self._records_wifi = records_wifi

    @property
    def records_bluetooth(self) -> Optional[RecordsBluetooth]:
        """
        :return: Bluetooth records
        """
        return self._records_bluetooth

    @records_bluetooth.setter
    def records_bluetooth(self, records_bluetooth: Optional[RecordsBluetooth]):
        if records_bluetooth is not None and not isinstance(records_bluetooth, RecordsBluetooth):
            raise TypeError('RecordsBluetooth expected')
        self._records_bluetooth = records_bluetooth

    @property
    def records_gnss(self) -> Optional[RecordsGnss]:
        """
        :return: GNSS records
        """
        return self._records_gnss

    @records_gnss.setter
    def records_gnss(self, records_gnss: Optional[RecordsGnss]):
        if records_gnss is not None and not isinstance(records_gnss, RecordsGnss):
            raise TypeError('RecordsGnss expected')
        self._records_gnss = records_gnss

    @property
    def records_accelerometer(self) -> Optional[RecordsAccelerometer]:
        """
        :return: Acceleration records
        """
        return self._records_accelerometer

    @records_accelerometer.setter
    def records_accelerometer(self, records_accelerometer: Optional[RecordsAccelerometer]):
        if records_accelerometer is not None and not isinstance(records_accelerometer, RecordsAccelerometer):
            raise TypeError('RecordsAccelerometer expected')
        self._records_accelerometer = records_accelerometer

    @property
    def records_gyroscope(self) -> Optional[RecordsGyroscope]:
        """
        :return: Gyroscope records
        """
        return self._records_gyroscope

    @records_gyroscope.setter
    def records_gyroscope(self, records_gyroscope: Optional[RecordsGyroscope]):
        if records_gyroscope is not None and not isinstance(records_gyroscope, RecordsGyroscope):
            raise TypeError('RecordsGyroscope expected')
        self._records_gyroscope = records_gyroscope

    @property
    def records_magnetic(self) -> Optional[RecordsMagnetic]:
        """
        :return: Magnetic records
        """
        return self._records_magnetic

    @records_magnetic.setter
    def records_magnetic(self, records_magnetic: Optional[RecordsMagnetic]):
        if records_magnetic is not None and not isinstance(records_magnetic, RecordsMagnetic):
            raise TypeError('RecordsMagnetic expected')
        self._records_magnetic = records_magnetic

    @property
    def keypoints(self) -> Optional[Dict[str, Keypoints]]:
        """
        :return: the keypoints collection
        """
        return self._keypoints

    @keypoints.setter
    def keypoints(self, keypoints: Optional[Dict[str, Keypoints]]):
        if keypoints is not None and not isinstance(keypoints, dict):
            raise TypeError('dict of Keypoints expected')
        self._keypoints = keypoints

    @property
    def descriptors(self) -> Optional[Dict[str, Descriptors]]:
        """
        :return: the descriptors collection
        """
        return self._descriptors

    @descriptors.setter
    def descriptors(self, descriptors: Optional[Dict[str, Descriptors]]):
        if descriptors is not None and not isinstance(descriptors, dict):
            raise TypeError('dict of Descriptors expected')
        self._descriptors = descriptors

    @property
    def global_features(self) -> Optional[Dict[str, GlobalFeatures]]:
        """
        :return: the global features collection
        """
        return self._global_features

    @global_features.setter
    def global_features(self, global_features: Optional[Dict[str, GlobalFeatures]]):
        if global_features is not None and not isinstance(global_features, dict):
            raise TypeError('dict of GlobalFeatures expected')
        self._global_features = global_features

    @property
    def matches(self) -> Optional[Dict[str, Matches]]:
        """
        :return: the matches collection
        """
        return self._matches

    @matches.setter
    def matches(self, matches: Optional[Dict[str, Matches]]):
        if matches is not None and not isinstance(matches, dict):
            raise TypeError('dict of Matches expected')
        self._matches = matches

    @property
    def observations(self) -> Optional[Observations]:
        """
        :return: the observations
        """
        return self._observations

    @observations.setter
    def observations(self, observations: Optional[Observations]):
        if observations is not None and not isinstance(observations, Observations):
            raise TypeError('Observations expected')
        self._observations = observations

    @property
    def points3d(self) -> Optional[Points3d]:
        """
        :return: the 3D points
        """
        return self._points3d

    @points3d.setter
    def points3d(self, points3d: Optional[Points3d]):
        if points3d is not None and not isinstance(points3d, Points3d):
            raise TypeError('Points3d expected')
        self._points3d = points3d

    def as_dict(self, keep_none=False) -> Dict[str, Union[
        Sensors,
        Rigs,
        Trajectories,
        RecordsCamera,
        RecordsDepth,
        RecordsLidar,
        RecordsWifi,
        RecordsBluetooth,
        RecordsGnss,
        RecordsAccelerometer,
        RecordsGyroscope,
        RecordsMagnetic,
        Optional[Dict[str, Keypoints]],
        Optional[Dict[str, Descriptors]],
        Optional[Dict[str, GlobalFeatures]],
        Optional[Dict[str, Matches]],
        Observations,
        Points3d
    ]]:
        """ convenience accessor to all members at once """
        return {
            name[1:]: member
            for name, member in self.__dict__.items()
            if not name.startswith('__') and (keep_none or member is not None)
        }

    def __repr__(self) -> str:
        representation = '\n'.join([
            f'{name:14} : {len(value):4}'
            for name, value in self.as_dict().items()
        ])

        if len(representation) == 0:
            representation = 'no data'
        return representation



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports a list of images in the kapture format
"""

import argparse
import logging
import os
import os.path as path
import PIL
from PIL import Image
from typing import List

# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
from kapture.io.records import TransferAction, import_record_data_from_dir_auto

logger = logging.getLogger('import_image_list')


def import_image_list(images_list_filenames: List[str],
                      images_dirpath: str,
                      kapture_path: str,
                      force_overwrite_existing: bool = False,
                      images_import_method: TransferAction = TransferAction.skip) -> None:
    """
    Imports the list of images to a kapture. This creates only images and cameras.

    :param images_list_filenames: list of text files containing image file names
    :param images_dirpath: path to images directory.
    :param kapture_path: path to kapture root directory.
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    """
    assert isinstance(images_list_filenames, list)
    os.makedirs(kapture_path, exist_ok=True)
    delete_existing_kapture_files(kapture_path, force_erase=force_overwrite_existing)

    cameras = kapture.Sensors()
    images = kapture.RecordsCamera()

    offset = 0
    logger.info('starting conversion...')
    for images_list_filename in images_list_filenames:
        logger.info(f'loading {images_list_filename}')
        with open(images_list_filename) as file:
            images_list = file.readlines()
            # remove end line char and empty lines
            images_list = [line.rstrip() for line in images_list if line != '\n']

            for i in range(0, len(images_list)):
                line = images_list[i].split()
                image_file_name = line[0]
                if len(line) > 1:
                    model = line[1]
                    model_params = line[2:]
                else:
                    model = kapture.CameraType.UNKNOWN_CAMERA.value
                    try:
                        # lazy open
                        with Image.open(path.join(images_dirpath, image_file_name)) as im:
                            width, height = im.size
                            model_params = [width, height]
                    except (OSError, PIL.UnidentifiedImageError):
                        # It is not a valid image: skip it
                        logger.info(f'Skipping invalid image file {image_file_name}')
                        continue

                camera_id = f'sensor{i + offset}'
                cameras[camera_id] = kapture.Camera(model, model_params)
                images[(i + offset, camera_id)] = image_file_name
            offset += len(images_list)

    # import (copy) image files.
    logger.info('import image files ...')
    filename_list = [f for _, _, f in kapture.flatten(images)]
    import_record_data_from_dir_auto(images_dirpath, kapture_path, filename_list, images_import_method)

    # pack into kapture format
    imported_kapture = kapture.Kapture(sensors=cameras, records_camera=images)
    logger.info('writing imported data...')
    kapture_to_dir(kapture_path, imported_kapture)


def import_image_list_command_line() -> None:
    """
    Do the image list import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='import image list text files to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '--force', action='store_true', default=False,
                        help='force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--inputs', nargs='+', required=True,
                        help=('input path to image list files:\n'
                              'image_name camera_type camera_params\n or\n image_name'))
    parser.add_argument('-o', '--output', required=True, help='output directory')
    parser.add_argument('-im', '--image_path', default="",
                        help='path to images. Only needed when inputs do not have camera parameters')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    ####################################################################################################################
    args = parser.parse_args()

    """
    image_name camera_type camera_params
    example: image_name SIMPLE_RADIAL w h f cx cy r
    or
    image_name
    """

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_image_list(
        images_list_filenames=args.inputs,
        images_dirpath=args.image_path,
        kapture_path=args.output,
        force_overwrite_existing=args.force,
        images_import_method=args.image_transfer
    )


if __name__ == '__main__':
    import_image_list_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to export a kapture into an openmvg file.
"""

import argparse
import logging
import sys
import os.path as path
# kapture
import path_to_kapture  # enables import kapture  # noqa: F401
import kapture.utils.logging
from kapture.io.records import TransferAction
from kapture.converter.openmvg.export_openmvg import export_openmvg

logger = logging.getLogger('openmvg')


def export_openmvg_command_line():
    """
    Do the kapture to openmvg export using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='Exports from kapture format to openMVG JSON file.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output file if already exists.')
    # create the parser for the export command #########################################################################
    parser.add_argument('-i', '-k', '--kapture', required=True,
                        help='path to input kapture data root directory')
    parser.add_argument('-o', '--openmvg',
                        help='root path to openmvg data structure (subpath are guessed).')
    parser.add_argument('-s', '--sfm_data',
                        help='path to output openMVG sfm_data JSON file.')
    parser.add_argument('-im', '--images',
                        help='path to output openMVG image directory.')
    parser.add_argument('-r', '--regions',
                        help='path to output openMVG regions directory (for features and descriptors).')
    parser.add_argument('-m', '--matches',
                        help='path to output openMVG matches file.')
    parser.add_argument('--image_action', default='skip', type=TransferAction,
                        help=f'''what to do with images:
        {TransferAction.root_link.name}: link to the root of the images directory (default) ;
        {TransferAction.link_absolute.name}: absolute individual file link ;
        {TransferAction.link_relative.name}: relative individual file link ;
        {TransferAction.copy.name}: copy file instead of creating link ;
        {TransferAction.move.name}: move file instead of creating link ;
        {TransferAction.skip.name}: do not create links
                                      ''')
    parser.add_argument('--image_path_flatten', action='store_true',
                        help='flatten image subpath, to make sure there is no collision in image names.')
    parser.add_argument('-kpt', '--keypoints-type', default=None, help='kapture keypoints type.')
    parser.add_argument('-desc', '--descriptors-type', default=None, help='kapture descriptors type.')
    args = parser.parse_args()
    logger.setLevel(args.verbose)

    if args.openmvg is None and args.sfm_data is None:
        raise ValueError('at least one output path should be given.')

    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)
    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    # auto complete using args.openmvg
    if args.openmvg and args.sfm_data is None:
        args.sfm_data = path.join(args.openmvg, 'sfm_data.json')
        logger.debug(f'guessing output sfm_data is {args.sfm_data}')
    if args.openmvg and args.images is None:
        args.images = path.join(args.openmvg, 'images')
        logger.debug(f'guessing output images is {args.images}')
    if args.openmvg and args.regions is None:
        args.regions = path.join(args.openmvg, 'matches')
        logger.debug(f'guessing output regions is {args.regions}')
    if args.openmvg and args.matches is None:
        args.matches = path.join(args.openmvg, 'matches', 'matches.txt')
        logger.debug(f'guessing output matches is {args.matches}')

    # no image dir == sip transfer
    if args.images is None:
        args.image_action = TransferAction.skip
    if args.image_action == TransferAction.skip:
        args.images = None

    if not args.image_action == TransferAction.skip:
        if args.images is None:
            raise ValueError(f'You must specify an image root path for this transfer ({args.image_action})')

    export_openmvg(
        kapture_path=args.kapture,
        openmvg_sfm_data_file_path=args.sfm_data,
        openmvg_image_root_path=args.images,
        openmvg_regions_dir_path=args.regions,
        openmvg_matches_file_path=args.matches,
        image_action=args.image_action,
        image_path_flatten=args.image_path_flatten,
        force=args.force,
        keypoints_type=args.keypoints_type,
        descriptors_type=args.descriptors_type
    )


if __name__ == '__main__':
    try:
        export_openmvg_command_line()
    except Exception:
        logger.error('Fatal error', exc_info=True)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script merges several kapture dataset to a new kapture. We might also optionally remove some of the data by type.
"""

import argparse
import logging
import os
# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from typing import List

from kapture.algo.merge_keep_ids import merge_keep_ids
from kapture.algo.merge_remap import merge_remap
from kapture.io.csv import get_all_tar_handlers, kapture_from_dir, kapture_to_dir
from kapture.io.records import get_image_fullpath
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.records import TransferAction


logger = logging.getLogger('merge')


def merge_kaptures(kapture_path_list: List[str],  # noqa: C901: function really not complex
                   merged_path: str,
                   keep_sensor_ids:  bool,
                   images_import_strategy: TransferAction = TransferAction.skip,
                   skip: List[str] = [],
                   force: bool = False) -> None:
    """
    Merge a list of kapture dataset to a new one.

    :param kapture_path_list: list of path to the top directory of the kapture datasets to merge
    :param merged_path: path to the merged top directory kapture to create
    :param keep_sensor_ids: if True, will keep the original sensor identifiers. Otherwise, might rename them.
    :param images_import_strategy: import action to apply on image files
    :param skip: list of kapture data type names to optionally skip (trajectories, records_camera, descriptors, ...)
    :param force: If True, silently overwrite kapture files if already exists.
    """
    os.makedirs(merged_path, exist_ok=True)
    delete_existing_kapture_files(merged_path, force_erase=force)

    skip_list = []
    if 'trajectories' in skip:
        skip_list.append(kapture.Trajectories)
    if 'records_camera' in skip:
        skip_list.append(kapture.RecordsCamera)
    if 'records_lidar' in skip:
        skip_list.append(kapture.RecordsLidar)
    if 'records_wifi' in skip:
        skip_list.append(kapture.RecordsWifi)
    if 'records_bluetooth' in skip:
        skip_list.append(kapture.RecordsBluetooth)
    if 'records_gnss' in skip:
        skip_list.append(kapture.RecordsGnss)
    if 'records_accelerometer' in skip:
        skip_list.append(kapture.RecordsAccelerometer)
    if 'records_gyroscope' in skip:
        skip_list.append(kapture.RecordsGyroscope)
    if 'records_magnetic' in skip:
        skip_list.append(kapture.RecordsMagnetic)
    if 'keypoints' in skip:
        skip_list.append(kapture.Keypoints)
    if 'descriptors' in skip:
        skip_list.append(kapture.Descriptors)
    if 'global_features' in skip:
        skip_list.append(kapture.GlobalFeatures)
    if 'matches' in skip:
        skip_list.append(kapture.Matches)
    if 'points3d' in skip:
        skip_list.append(kapture.Points3d)
    if 'observations' in skip:
        skip_list.append(kapture.Observations)

    kapture_data_list = []
    kapture_tarcollection_list = []
    try:
        for kapture_path in kapture_path_list:
            logger.info(f'Loading {kapture_path}')
            tar_handlers = get_all_tar_handlers(kapture_path)
            kapture_data = kapture_from_dir(kapture_path, tar_handlers=tar_handlers)
            kapture_data_list.append(kapture_data)
            kapture_tarcollection_list.append(tar_handlers)

        if keep_sensor_ids:
            merged_kapture = merge_keep_ids(kapture_data_list, skip_list,
                                            kapture_path_list, kapture_tarcollection_list,
                                            merged_path, images_import_strategy)
        else:
            merged_kapture = merge_remap(kapture_data_list, skip_list,
                                         kapture_path_list, kapture_tarcollection_list,
                                         merged_path, images_import_strategy)
    finally:
        for tar_handlers in kapture_tarcollection_list:
            tar_handlers.close()

    logger.info('Writing merged kapture data...')
    kapture_to_dir(merged_path, merged_kapture)


def merge_command_line() -> None:
    """
    Do the merge_kaptures of several kapture datasets using the command line parameters provided by the user.
    """
    parser = argparse.ArgumentParser(
        description=(f'merge multiple data in the kapture pivot format. Warning: the folder {get_image_fullpath("")} '
                     'will not be copied, remember to manually copy/create a symlink to it.'
                     ' Warning n°2: the inputs records should be disjoint.'))
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-i', '--inputs', nargs='+', required=True,
                        help='input paths to kapture data')
    parser.add_argument('-o', '--output', required=True,
                        help='output directory for the merged data')
    parser.add_argument('--keep-sensor-ids', action='store_true', default=False,
                        help='do not remap sensor ids. if duplicates are found, the first one will be kept')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.skip,
                        help=f'How to import images [skip], '
                        f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output directory if already exists.')
    parser.add_argument('-s', '--skip',
                        choices=['trajectories', 'records_camera', 'records_lidar', 'records_wifi',
                                 'keypoints', 'descriptors', 'global_features',
                                 'matches', 'points3d', 'observations'],
                        nargs='+', default=[], help='data to skip')

    args = parser.parse_args()
    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)
    merge_kaptures(args.inputs, args.output, args.keep_sensor_ids, args.image_transfer, args.skip, args.force)


if __name__ == '__main__':
    merge_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Virtual gallery to kapture import functions.
"""

import os
import logging
from typing import List
# kapture
import kapture
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
# local
from .virtual_gallery_testing import import_testing_intrinsics, import_testing_extrinsics
from .virtual_gallery_testing import convert_testing_intrinsics, convert_testing_extrinsics
from .virtual_gallery_training import import_training_intrinsics, import_training_extrinsics
from .virtual_gallery_training import convert_training_intrinsics, convert_training_extrinsics
from .virtual_gallery_training import training_rig_config


logger = logging.getLogger('virtual_gallery')


def import_virtual_gallery(input_root_path: str,
                           configuration: str,
                           light_range: List[int],
                           loop_range: List[int],
                           camera_range: List[int],
                           occlusion_range: List[int],
                           as_rig: bool,
                           images_import_method: TransferAction,
                           kapture_path: str,
                           force_overwrite_existing: bool = False) -> None:
    """
    Creates a kapture with a virtual gallery.

    :param input_root_path: root path of virtual gallery
    :param configuration: training, testing or all (both)
    :param light_range: list of lights to include
    :param loop_range: list of training loops to include
    :param camera_range: list of training cameras to include
    :param occlusion_range: list of testing occlusion levels to include
    :param as_rig: in training trajectories, writes the position of the rig instead of individual cameras
    :param kapture_path: path to kapture top directory
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    """
    # Check for existing files
    os.makedirs(kapture_path, exist_ok=True)
    delete_existing_kapture_files(kapture_path, force_overwrite_existing)

    offset = 0
    cameras = kapture.Sensors()
    images = kapture.RecordsCamera()
    trajectories = kapture.Trajectories()
    rigs = kapture.Rigs()

    # Process all training data
    if configuration == "training" or configuration == "all":
        logger.info("Reading training files")
        camera_range_set = set(camera_range)
        training_intrinsics = import_training_intrinsics(input_root_path, light_range, loop_range, camera_range_set)
        training_extrinsics = import_training_extrinsics(input_root_path, light_range, loop_range, camera_range_set)

        convert_training_intrinsics(training_intrinsics, cameras)
        convert_training_extrinsics(offset, training_extrinsics, images, trajectories, as_rig)
        rigs.update(training_rig_config)

        offset += len(training_extrinsics)
    # Process all testing data
    if configuration == "testing" or configuration == "all":
        logger.info("Reading testing files")
        testing_intrinsics = import_testing_intrinsics(input_root_path, light_range, occlusion_range)
        testing_extrinsics = import_testing_extrinsics(input_root_path, light_range, occlusion_range)

        convert_testing_intrinsics(testing_intrinsics, cameras)
        convert_testing_extrinsics(offset, testing_extrinsics, images, trajectories)

        offset += len(testing_extrinsics)

    logger.info("Writing imported data to disk")
    kapture_data = kapture.Kapture(sensors=cameras, records_camera=images, trajectories=trajectories, rigs=rigs or None)
    # import images
    image_list = [name for _, _, name in kapture.flatten(kapture_data.records_camera)]
    import_record_data_from_dir_auto(input_root_path, kapture_path, image_list, images_import_method)
    kapture_to_dir(kapture_path, kapture_data)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Colmap specific database functions
"""

from kapture.io.tar import TarCollection
import logging
import numpy as np
import os
import os.path as path
from typing import List, Tuple, Dict, Optional, Set
from tqdm import tqdm

import sqlite3.dbapi2

# kapture
import kapture
from kapture.io.features import image_keypoints_from_file

# local
from .database import COLMAPDatabase, image_ids_to_pair_id, pair_id_to_image_ids
from .cameras import get_colmap_camera
from .export_colmap_reconstruction import export_to_colmap_txt

logger = logging.getLogger('colmap')


def foreign_keys_off(database: COLMAPDatabase) -> sqlite3.dbapi2.Cursor:
    """
    Set foreign keys off.
     Warning: don't forget to set it "on" afterwards!

    :param database: database to execute the command against
    :return: database cursor
    """
    return database.execute("PRAGMA foreign_keys = off;")


def foreign_keys_on(database: COLMAPDatabase) -> sqlite3.dbapi2.Cursor:
    """
    Set foreign keys on.

    :param database: database to execute the command against
    :return: database cursor
    """
    return database.execute("PRAGMA foreign_keys = on;")


def get_camera_ids_from_database(database: COLMAPDatabase) -> List[int]:
    """
    Get the list of colmap camera ids

    :param database: colmap database
    :return: list of colmap camera ids
    """
    return [camera_id for camera_id, in database.execute('SELECT camera_id FROM cameras;')]


def get_images_from_database(database: COLMAPDatabase) -> List[Tuple[str, int]]:
    """
    Get the list (image_name, camera_id) for the database

    :param database: colmap database
    :return: list of tuple (image_name, camera_id)
    """
    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    image_list = []
    for name, camera_id \
            in tqdm(database.execute('SELECT name, camera_id FROM images;'), disable=hide_progressbar):
        image_list.append((name, camera_id))
    return image_list


def get_keypoints_set_from_database(database: COLMAPDatabase,
                                    kapture_image_name_from_colmap_id: Dict[int, str]) -> Set[str]:
    """
    return set(image_name for image_name in colmap db keypoint table)

    :param database: colmap database
    :type database: COLMAPDatabase
    :param kapture_image_name_from_colmap_id: dict colmap image_id -> kapture image_name
    :type kapture_image_name_from_colmap_id: Dict[int, str]
    """
    return set(kapture_image_name_from_colmap_id[image_id]
               for (image_id,) in database.execute("SELECT image_id FROM keypoints"))


def get_matches_set_from_database(database: COLMAPDatabase,
                                  kapture_image_name_from_colmap_id: Dict[int, str]) -> Set[Tuple[str, str]]:
    """
    return set(pair_image_name for pair_image_name in colmap db matches table)

    :param database: colmap database
    :type database: COLMAPDatabase
    :param kapture_image_name_from_colmap_id: dict colmap image_id -> kapture image_name
    :type kapture_image_name_from_colmap_id: Dict[int, str]
    """
    colmap_matches = (pair_id_to_image_ids(pair_id)
                      for (pair_id,) in database.execute('SELECT pair_id FROM matches'))
    return set((kapture_image_name_from_colmap_id[pair[0]], kapture_image_name_from_colmap_id[pair[1]])
               for pair in colmap_matches)


def update_DB_cameras_and_poses(database: COLMAPDatabase,
                                kapture_data: kapture.Kapture) -> None:
    """
    Update the colmap database with cameras and pose from a kapture.

    :param database: input/output colmap database.
    :param kapture_data: kapture to read data from
    """
    if kapture_data.rigs:
        raise ValueError('update_DB_cameras_and_poses does not handle rigs, please use rigs_remove_inplace before.')

    # 1 - Remove foreign key constraints
    # 2 - empty camera table
    # 3 - insert new cameras with intrinsics
    # 4 - update images to use new camera IDs, and add pose from trajectories
    # 5 - Re-add foreign key constraints

    # WARNING: Delete all existing cameras
    logger.debug('Delete existing cameras in database...')
    foreign_keys_off(database)
    database.execute("DELETE from cameras")

    logger.debug('Add cameras in database...')
    colmap_camera_ids = {}  # to keep ID of camera in case it is different from the one we use
    for cam_id, cam in kapture_data.sensors.items():
        if cam.sensor_type not in kapture.ALL_CAMERA_SENSOR_TYPES:
            continue
        col_cam_id, width, height, params, prior_focal_length = get_colmap_camera(cam)

        colmap_camera_ids[cam_id] = database.add_camera(col_cam_id,
                                                        # image size
                                                        width, height,
                                                        params,
                                                        prior_focal_length=prior_focal_length)

    logger.debug('Update images table with new camera ID and pose from trajectories')
    # compute actual camera trajectories from rig trajectories

    images_cam_ts = {image_path: (ts, cam_id)
                     for ts, shot in kapture_data.records_camera.items()
                     for cam_id, image_path in shot.items()}
    logger.info('register images in database...')
    for name, (timestamp, cam_id) in images_cam_ts.items():
        # retrieve image pose from trajectories
        if timestamp not in kapture_data.trajectories:
            # no pose for that timestamp # TODO what should we do ?
            prior_q = [0.0] * 4
            prior_t = [0.0] * 3
        else:
            assert cam_id in kapture_data.trajectories[timestamp]
            pose_tr = kapture_data.trajectories[timestamp].get(cam_id)
            prior_q = pose_tr.r_raw
            prior_t = pose_tr.t_raw

        # Update image in DB
        update_image(database, name, colmap_camera_ids[cam_id], prior_q=prior_q, prior_t=prior_t)

    # Foreign key constraints should be OK now
    foreign_keys_on(database)
    database.commit()


def remove_camera(database: COLMAPDatabase, camera_id: int) -> None:
    """
    Removes a camera from the colmap database (from cameras table only).

    :param database: input/output colmap database.
    :param camera_id: identifier of camera
    """
    try:
        database.execute("DELETE FROM cameras WHERE camera_id = ?", (camera_id,))
    except Exception as e:
        logger.warning(e)
        pass


def update_image(database: COLMAPDatabase,
                 name,
                 camera_id,
                 prior_q=np.zeros(4),
                 prior_t=np.zeros(3)) -> None:
    """
    Update an image in the colmap database

    :param database: colmap database to update
    :param name: image name
    :param camera_id: camera identifier
    :param prior_q:
    :param prior_t:
    """
    try:
        database.execute(
            "UPDATE images SET camera_id = ?, prior_qw = ?, prior_qx = ?, prior_qy = ?, prior_qz = ?, prior_tx = ?,"
            " prior_ty = ?, prior_tz = ? WHERE name = ?",
            (camera_id, prior_q[0], prior_q[1], prior_q[2], prior_q[3], prior_t[0], prior_t[1], prior_t[2], name))
    except Exception as e:
        logger.warning(e)
        pass


def get_colmap_camera_ids_from_db(database: COLMAPDatabase, images: kapture.RecordsCamera) -> Dict[str, int]:
    """
    returns a dict mapping colmap_camera_ID -> sensor_id

    :param database: input colmap database.
    :param images: input kapture recorded images.
    :return: a dict of colmap camera ids kapture_id -> colmap_id
    """
    # colmap_camera_ID = {}  # colmap_camera_ID[sensor_id] = ids
    colmap_image_to_camera = {
        name: camera_id
        for camera_id, name in database.execute("SELECT camera_id, name FROM images")
    }
    colmap_camera_ids = {}
    for timestamp, sensor_id, sensing_filepath in kapture.flatten(images):
        if sensor_id not in colmap_camera_ids:
            colmap_camera_id = colmap_image_to_camera[sensing_filepath]
            colmap_camera_ids[sensor_id] = colmap_camera_id

    return colmap_camera_ids


def get_colmap_image_ids_from_db(database: COLMAPDatabase):
    """
    returns a dict mapping image_name -> colmap_image_ID

    :param database: input colmap database to write in.
    :return: a dict of image_name -> colmap_image_ID
    """
    # colmap_image_ids = {}  # colmap_image_ids[name] = ids
    return {
        name: image_id
        for image_id, name in database.execute("SELECT image_id, name FROM images")
    }


def is_colmap_db_empty(database: COLMAPDatabase) -> bool:
    """
    Check if the given colmap database is empty, ie. all colmap table are
        - are not there,
        - or empty

    :param database: input colmap database to write in.
    :return: True if the given colmap database is empty
    """
    # check database is empty
    table_names = ['cameras', 'descriptors', 'images', 'keypoints',
                   'matches', 'sqlite_sequence', 'two_view_geometries']
    for table in table_names:
        try:
            cursor = database.execute("SELECT count(*) FROM {}".format(table))
            count = cursor.fetchall()[0][0]
            if count > 0:
                return False
        except Exception:
            continue
    return True


def exists_table(table: str, database: COLMAPDatabase) -> bool:
    """
    Check if a table exist in a database

    :param table: name of table to check
    :param database: colmap database.
    :return: True if the given table exists
    """
    try:
        cursor = database.execute("SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{}'".format(table))
        count = cursor.fetchall()[0][0]
        if count == 0:
            return False
    except Exception:
        return False

    return True


def save_match_list(kapture_matches: kapture.Matches, path_file: str) -> None:
    """
    Saves list of matching files from kapture data to simple text file
    format :
    image_match_1 image_match_2

    :param kapture_matches: kapture_matches data
    :param kapture_matches: path to the matches file

    """
    if path.exists(path_file):
        os.remove(path_file)
    f = open(path_file, "w")
    for image_path1, image_path2 in kapture_matches:
        f.write("{} {}\n".format(image_path1, image_path2))
    f.close()


def generate_priors_for_reconstruction(kapture_data: kapture.Kapture,
                                       database: COLMAPDatabase,
                                       path_to_priors_for_reconstruction: str) -> None:
    """
    Generate priors for a reconstruction using cameras, images and trajectories.

    :param kapture_data: kapture data
    :param database: colmap database.
    :param path_to_priors_for_reconstruction: path to the priors file
    """
    colmap_camera_ids = get_colmap_camera_ids_from_db(database, kapture_data.records_camera)
    colmap_image_ids = get_colmap_image_ids_from_db(database)
    kapture_data_copy = kapture.Kapture(sensors=kapture_data.sensors,
                                        records_camera=kapture_data.records_camera,
                                        trajectories=kapture_data.trajectories,
                                        rigs=kapture_data.rigs)
    # in priors, do not copy keypoints, points3d
    export_to_colmap_txt(
        path_to_priors_for_reconstruction,
        kapture_data_copy,
        "",  # kapture_data_copy do not have binaries so path is irrelevant
        None,  # same for tar as above
        colmap_camera_ids,
        colmap_image_ids)


def add_cameras_to_database(sensors: kapture.Sensors, database: COLMAPDatabase) -> Dict[str, int]:
    """
    Add the kapture cameras to the colmap database.

    :param sensors: list af kapture cameras
    :param database: colmap database.
    :return: dictionary of kapture camera identifier -> colmap camera identifier
    """
    assert isinstance(sensors, kapture.Sensors)
    colmap_camera_ids = {}
    cameras = {cam_id: camera
               for cam_id, camera in sensors.items()
               if isinstance(camera, kapture.Camera)}

    for cam_id, cam in cameras.items():
        col_cam_id, width, height, params, prior_focal_length = get_colmap_camera(cam)
        colmap_camera_ids[cam_id] = database.add_camera(
            col_cam_id,
            width,
            height,
            params,
            prior_focal_length=prior_focal_length)

    database.commit()
    return colmap_camera_ids


def get_images_as_list_in_colmap_format(flatten_images: List[Tuple[int, str, str]],
                                        trajectories: Optional[kapture.Trajectories],
                                        colmap_camera_ids: Dict[str, int]
                                        ) -> List[Tuple[str, int, List[float], List[float]]]:
    """
    Returns data formatted to fill the colmap "images" table in the database.

    :param flatten_images: input list of images obtained through flatten(records_camera)
    :param trajectories: input optional kapture trajectories. Poses are set to 0 if not given.
    :param colmap_camera_ids: input dict mapping kapture sensor ids to colmap camera ids.
    :return: A list of tuple containing data to fill the colmap "images" table, ie
            image_filename, colmap_cam_id, rotation, translation
    """
    images_in_colmap_format = []
    for timestamp, sensor_id, image_filename in flatten_images:
        # retrieve image pose from trajectories
        prior_q, prior_t = None, None
        if trajectories is not None and (timestamp, sensor_id) in trajectories:
            # there is a pose for that timestamp / sensor_id pair
            pose_tr = trajectories[(timestamp, sensor_id)]
            prior_q = pose_tr.r_raw
            prior_t = pose_tr.t_raw
        # set fall back values if pose is undefined
        if prior_q is None:
            prior_q = 4 * [0.0]
        if prior_t is None:
            prior_t = 3 * [0.0]
        images_in_colmap_format.append(
            (image_filename, colmap_camera_ids[sensor_id], prior_q, prior_t)
        )
    # images_in_colmap_format = image_filename, colmap_cam_id, rotation, translation
    return images_in_colmap_format


def update_images_from_list_in_colmap_format(database: COLMAPDatabase,
                                             image_list: List[Tuple[str, str, List[float], List[float]]]) -> None:
    """
    Update images in the colmap database.

    :param database: colmap database.
    :param image_list: list of images information to update
    """
    for name, cam_id, prior_q, prior_t in image_list:
        update_image(database,
                     name, cam_id,
                     prior_q=prior_q,
                     prior_t=prior_t)
    database.commit()


def add_images_from_list_in_colmap_format(database: COLMAPDatabase,
                                          image_list: List[Tuple[str, str, List[float], List[float]]]
                                          ) -> Dict[str, int]:
    """
    Add images to the colmap database.

    :param database: input/output colmap database.
    :param image_list: list of image information
    :return: dict mapping kapture image ids to colmap image ids.
    """
    colmap_image_ids = {}  # colmap_image_ids[name] = ids
    for name, cam_id, prior_q, prior_t in image_list:
        colmap_image_ids[name] = database.add_image(
            name, cam_id,
            prior_q=prior_q,
            prior_t=prior_t)
    database.commit()
    return colmap_image_ids


def update_images_in_database_from_flatten(database: COLMAPDatabase,
                                           flatten_images: List[Tuple[int, str, str]],
                                           trajectories: kapture.Trajectories,
                                           colmap_camera_ids: dict) -> None:
    """
    Updates images information in colmap database.

    :param database: colmap database.
    :param flatten_images: list of image information
    :param trajectories: trajectories for the images
    :param colmap_camera_ids: kapture camera identifier -> colmap camera identifier dictionary
    """
    images_in_colmap_format = get_images_as_list_in_colmap_format(flatten_images, trajectories, colmap_camera_ids)
    update_images_from_list_in_colmap_format(database, images_in_colmap_format)


def add_images_to_database_from_flatten(database: COLMAPDatabase,
                                        flatten_images: List[Tuple[int, str, str]],
                                        trajectories: kapture.Trajectories,
                                        colmap_camera_ids: dict) -> Dict[str, int]:
    """
    Add images in the colmap database.

    :param database: colmap database.
    :param flatten_images: kapture images
    :param trajectories: images trajectories
    :param colmap_camera_ids: kapture camera identifier -> colmap camera identifier dictionary
    :return: dict mapping kapture image ids to colmap image ids.
    """
    images_in_colmap_format = get_images_as_list_in_colmap_format(flatten_images, trajectories, colmap_camera_ids)
    colmap_image_ids = add_images_from_list_in_colmap_format(database, images_in_colmap_format)
    return colmap_image_ids


def update_images_in_database(database: COLMAPDatabase,
                              images,
                              trajectories: kapture.Trajectories,
                              colmap_camera_ids: dict) -> None:
    """
    Update image information in colmap database.

    :param database: colmap database.
    :param images: kapture images
    :param trajectories: images trajectories
    :param colmap_camera_ids: kapture camera identifier -> colmap camera identifier dictionary
    """
    images_flattened = list(images.flattened())
    images_in_colmap_format = get_images_as_list_in_colmap_format(images_flattened, trajectories, colmap_camera_ids)
    update_images_from_list_in_colmap_format(database, images_in_colmap_format)


def add_images_to_database(database: COLMAPDatabase,
                           images: kapture.RecordsCamera,
                           trajectories: Optional[kapture.Trajectories],
                           colmap_camera_ids: Dict[str, int]
                           ) -> Dict[str, int]:
    """
    Add images in colmap database.

    :param database: colmap database.
    :param images: kapture images
    :param trajectories: images trajectories
    :param colmap_camera_ids: dict mapping kapture sensor ids to colmap camera ids.
    :return: dict mapping kapture image ids to colmap image ids.
    """
    assert isinstance(images, kapture.RecordsCamera)
    assert trajectories is None or isinstance(trajectories, kapture.Trajectories)
    images_flattened = list(kapture.flatten(images))
    images_in_colmap_format = get_images_as_list_in_colmap_format(images_flattened, trajectories, colmap_camera_ids)
    colmap_image_ids = add_images_from_list_in_colmap_format(database, images_in_colmap_format)
    return colmap_image_ids


def add_keypoints_to_database(database: COLMAPDatabase,
                              keypoints: kapture.Keypoints,
                              keypoints_type: str,
                              kapture_dir_path: str,
                              tar_handler: Optional[TarCollection],
                              colmap_image_ids: dict) -> None:
    """
    Add keypoints to the colmap database.

    :param database: colmap database.
    :param keypoints: kapture keypoints to add
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dir_path: kapture data top directory
    :param tar_handler: collection of preloaded tar archives
    :param colmap_image_ids: kapture camera identifier -> colmap camera identifier dictionary
    """
    keypoints_filepaths = kapture.io.features.keypoints_to_filepaths(keypoints, keypoints_type,
                                                                     kapture_dir_path, tar_handler)
    for image_filename, keypoints_filepath in keypoints_filepaths.items():
        image_keypoints = image_keypoints_from_file(keypoints_filepath, keypoints.dtype, keypoints.dsize)
        colmap_image_id = colmap_image_ids[image_filename]
        # Make sure keypoints are np.float32 and support by colmap
        if image_keypoints.shape[1] not in {2, 4, 6}:
            image_keypoints = image_keypoints[:, 0:2]
        image_keypoints = image_keypoints.astype(np.float32)
        database.add_keypoints(colmap_image_id, image_keypoints)
    database.commit()


def add_descriptors_to_database(database: COLMAPDatabase,
                                descriptors: kapture.Descriptors,
                                descriptors_type: str,
                                kapture_dir_path: str,
                                tar_handler: Optional[TarCollection],
                                colmap_image_ids: dict) -> None:
    """
    Add descriptors to the colmap database.

    :param database: colmap database.
    :param descriptors: kapture descriptors to add
    :param kapture_dir_path: kapture data top directory
    :param tar_handler: collection of preloaded tar archives
    :param colmap_image_ids: kapture camera identifier -> colmap camera identifier dictionary
    """
    descriptors_filepaths = kapture.io.features.descriptors_to_filepaths(descriptors,
                                                                         descriptors_type,
                                                                         kapture_dir_path,
                                                                         tar_handler)
    for image_filename, descriptors_filepath in descriptors_filepaths.items():
        image_descriptors = image_keypoints_from_file(descriptors_filepath, descriptors.dtype, descriptors.dsize)
        colmap_image_id = colmap_image_ids[image_filename]
        database.add_descriptors(colmap_image_id, image_descriptors)
    database.commit()


def add_matches_to_database(database: COLMAPDatabase,
                            matches: kapture.Matches,
                            keypoints_type: str,
                            kapture_dir_path: str,
                            tar_handler: Optional[TarCollection],
                            colmap_image_ids: dict,
                            export_two_view_geometry: bool = False) -> None:
    """
    Add matches to the colmap database.

    :param database: colmap database.
    :param matches: kapture matches to add
    :param kapture_dir_path: kapture data top directory
    :param tar_handler: collection of preloaded tar archives
    :param colmap_image_ids: kapture camera identifier -> colmap camera identifier dictionary
    :param export_two_view_geometry: if True, also export two geometry.
    """
    colmap_pairs_id = {}

    # matches[(image_path1, image_path2)] = image_matches
    matches.normalize()
    matches_filepaths = kapture.io.features.matches_to_filepaths(matches, keypoints_type, kapture_dir_path, tar_handler)
    for (image_path1, image_path2), image_matches_filepath in matches_filepaths.items():
        image_matches = kapture.io.features.image_matches_from_file(image_matches_filepath)
        colmap_image_id1 = colmap_image_ids[image_path1]
        colmap_image_id2 = colmap_image_ids[image_path2]
        colmap_pair_id = image_ids_to_pair_id(colmap_image_id1, colmap_image_id2)
        if colmap_pair_id in colmap_pairs_id:
            logging.warning('{} IS ALREADY IN DATABASE ({}, {})'.format(
                colmap_pair_id, *colmap_pairs_id[colmap_pair_id]))
            continue
        colmap_pairs_id[colmap_pair_id] = (image_path1, image_path2)
        # convert kapture to colmap matches (drop the score col, and convert to int)
        image_matches = image_matches[:, :-1].astype(np.uint32)
        try:
            database.add_matches(colmap_image_id1, colmap_image_id2, image_matches)
            if export_two_view_geometry:
                database.add_two_view_geometry(colmap_image_id1, colmap_image_id2, image_matches)
        except Exception as err:
            logger.warning(f'({image_path1}, {image_path2}) failed: {err}')

    database.commit()


def kapture_to_colmap(kapture_data: kapture.Kapture,
                      kapture_dirpath: str,
                      tar_handler: Optional[TarCollection],
                      database: COLMAPDatabase,
                      keypoints_type: Optional[str] = None,
                      descriptors_type: Optional[str] = None,
                      export_two_view_geometry: bool = False) -> None:
    """
    Export kapture data to colmap database.

    :param kapture_data: kapture data to export
    :param kapture_dirpath: path to kapture directory, to retrieve binary files (keypoints, descriptors, ...)
    :param tar_handler: collection of preloaded tar archives
    :param database: colmap database.
    :param keypoints_type: type of keypoints to export, name of the keypoints subfolder
    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
    :param export_two_view_geometry: if True, also export two geometry.
    """
    assert isinstance(kapture_data, kapture.Kapture)
    assert kapture_data.sensors is not None
    assert kapture_data.records_camera is not None

    if not is_colmap_db_empty(database):
        raise ValueError('the existing colmap database is not empty : {}'.format(database))

    # compute actual camera trajectories from rigs trajectories
    if kapture_data.rigs:
        # check the rigs are not used in trajectories
        used_sensors = set(sensor_id
                           for shot in kapture_data.trajectories.values()
                           for sensor_id in shot.keys())
        rigs_ids = set(rig_id for rig_id in kapture_data.rigs.keys())
        if any(used_sensor in rigs_ids for used_sensor in used_sensors):
            raise ValueError(
                'colmap format does not handle rigs notation. '
                'Remove rig from trajectories beforehand (see rigs_remove_inplace)')

    database.create_tables()

    # cameras
    logger.info(f'registering {len(kapture_data.sensors)} sensors (cameras) in database...')
    colmap_camera_ids = add_cameras_to_database(kapture_data.sensors, database)

    # images
    logger.info(f'registering {len(list(kapture.flatten(kapture_data.records_camera)))} images in database...')
    colmap_image_ids = add_images_to_database(
        database, kapture_data.records_camera,
        kapture_data.trajectories, colmap_camera_ids)

    # keypoints
    if kapture_data.keypoints is not None and keypoints_type is not None and keypoints_type in kapture_data.keypoints:
        keypoints = kapture_data.keypoints[keypoints_type]
        logger.info(f'registering {len(keypoints)} keypoints in database...')
        add_keypoints_to_database(database, keypoints, keypoints_type,
                                  kapture_dirpath, tar_handler, colmap_image_ids)

    # descriptors
    if kapture_data.descriptors is not None \
            and descriptors_type is not None \
            and descriptors_type in kapture_data.descriptors:
        descriptors = kapture_data.descriptors[descriptors_type]
        logger.info(f'registering {len(descriptors)} descriptors in database...')
        add_descriptors_to_database(database, descriptors,
                                    descriptors_type, kapture_dirpath,
                                    tar_handler, colmap_image_ids)

    # matches
    if kapture_data.matches is not None \
            and keypoints_type is not None \
            and keypoints_type in kapture_data.matches:
        matches = kapture_data.matches[keypoints_type]
        logger.info(f'registering {len(matches)} matches in database...')
        add_matches_to_database(database, matches, keypoints_type,
                                kapture_dirpath, tar_handler,
                                colmap_image_ids, export_two_view_geometry)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Functions to import the rig part of colmap (json format).
"""

import json
import logging
from typing import Optional, Tuple
# kapture
import kapture
# local
from .cameras import get_camera_kapture_id_from_colmap_id


logger = logging.getLogger('colmap')


def import_colmap_rig_json(rigs_colmap: list,
                           images: Optional[kapture.RecordsCamera] = None,
                           trajectories: Optional[kapture.Trajectories] = None
                           ) -> Tuple[kapture.Rigs, kapture.RecordsCamera, Optional[kapture.Trajectories]]:
    """
    Build a kapture rig from colmap json file.

    :param rigs_colmap: colmap data describing the rig.
    :param images: input/output camera recordings: timestamps are modified to match
    :param trajectories: input/output trajectories: timestamps are modified to match
    :return: rigs, images and trajectories
    """
    rigs_kapture = kapture.Rigs()
    # camera_id (kapture) -> file prefix
    camera_prefixes = {}

    """ rigs_colmap
    [{
        "cameras": [
            {"camera_id": 1, "image_prefix": "leftraw/"},
            {"camera_id": 2, "image_prefix": "rightraw/"}
        ],
        "ref_camera_id": 1
    }]
    """

    for rig_idx_colmap, rig_colmap in enumerate(rigs_colmap):
        rig_id_kapture = f'rig{rig_idx_colmap}'  # make up a rig ID from its idx in colmap.
        for cam_colmap in rig_colmap['cameras']:
            # colmap_cam_id -> kapture_cam_id
            camera_id_colmap = cam_colmap['camera_id']
            camera_id_kapture = get_camera_kapture_id_from_colmap_id(camera_id_colmap)
            camera_prefixes[camera_id_kapture] = cam_colmap['image_prefix']
            # colmap does not store rig geometry, but only the fact there is one.
            pose_unknown = kapture.PoseTransform(r=None, t=None)
            rigs_kapture[rig_id_kapture, camera_id_kapture] = pose_unknown

    reconstructed_images = None
    reconstructed_trajectories = None
    if images:
        # image_filepath => (prefix, suffix)
        filepath_to_split_fix = {}
        # if there are images, modify their timestamps to match
        # first pass: gather actual timestamps from suffix
        # camera_suffixes = set()
        for timestamp, camera_id_kapture, image_filepath in kapture.flatten(images):
            if camera_id_kapture not in camera_prefixes:
                raise KeyError('unknown camera_id {}'.format(camera_id_kapture))
            camera_prefix = camera_prefixes[camera_id_kapture]
            if not image_filepath.startswith(camera_prefix):
                raise ValueError('inconsistent camera name')
            filepath_to_split_fix[image_filepath] = (
                image_filepath[0:len(camera_prefix)],
                image_filepath[len(camera_prefix):]
            )

        suffixes = sorted(set(suf for _, suf in filepath_to_split_fix.values()))
        suffix_to_timestamp = {suffix: idx for idx, suffix in enumerate(suffixes)}
        idx_to_timestamp = {
            colmap_idx: suffix_to_timestamp[filepath_to_split_fix[filepath][1]]
            for colmap_idx, _, filepath in kapture.flatten(images)
        }

        # second pass: reconstruct images with timestamp (frame number) instead of colmap idx
        reconstructed_images = kapture.RecordsCamera()
        for colmap_idx, camera_id_kapture, image_filepath in kapture.flatten(images):
            timestamp = idx_to_timestamp[colmap_idx]
            reconstructed_images[timestamp, camera_id_kapture] = image_filepath

        # third pass: [optional] reconstruct trajectories
        if trajectories:
            reconstructed_trajectories = kapture.Trajectories()
            for colmap_idx, camera_id_kapture, pose in kapture.flatten(trajectories):
                timestamp = idx_to_timestamp[colmap_idx]
                reconstructed_trajectories[timestamp, camera_id_kapture] = pose

    return rigs_kapture, reconstructed_images, reconstructed_trajectories


def import_colmap_rig(json_filepath: str,
                      images: kapture.RecordsCamera = None,
                      trajectories: kapture.Trajectories = None):
    """
    Build a kapture rig from colmap json file.

    :param json_filepath: colmap reconstruction json file path
    :param images: input/output camera recordings: timestamps are modified to match
    :param trajectories: input/output trajectories: timestamps are modified to match
    :return: rigs, images and trajectories
    """
    with open(json_filepath) as infile:
        rigs_colmap = json.load(infile)
    return import_colmap_rig_json(rigs_colmap, images, trajectories)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from typing import List
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.utils.logging import getLogger


def merge_records_data(image_list: List[List[str]],
                       image_paths: List[str],
                       kapture_path: str,
                       images_import_method: TransferAction):
    """
    Merge several records data. keep only the first image.

    :param image_list: list of image_names
    :param image_paths: list of image paths
    :param kapture_path: root path to the merged kapture
    :param images_import_method: choose how to import actual image files
    """
    assert len(image_list) > 0
    assert len(image_list) == len(image_paths)

    added_images = set()
    for images_filenames, record_dirpath in zip(image_list, image_paths):
        images_filenames_to_add = {images_filename
                                   for images_filename in images_filenames
                                   if images_filename not in added_images}
        import_record_data_from_dir_auto(record_dirpath, kapture_path, images_filenames_to_add, images_import_method)
        diff = set(images_filenames).difference(images_filenames_to_add)
        if len(diff) > 0:
            getLogger().warning(f'Cannot import some images because they were already added: {diff}')
        added_images.update(images_filenames_to_add)
